"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#ObjectMemory
	instanceVariableNames:''
	classVariableNames:'InternalErrorHandler UserInterruptHandler TimerInterruptHandler
		SpyInterruptHandler StepInterruptHandler
		ExceptionInterruptHandler ErrorInterruptHandler
		MemoryInterruptHandler SignalInterruptHandler
		ChildSignalInterruptHandler DisposeInterruptHandler
		RecursionInterruptHandler IOInterruptHandler
		CustomInterruptHandler RegisteredErrorInterruptHandlers
		InterruptLatencyMonitor AllocationFailureSignal
		MallocFailureSignal LowSpaceSemaphore IncrementalGCLimit
		FreeSpaceGCLimit FreeSpaceGCAmount BackgroundCollectProcess
		BackgroundFinalizationProcess FinalizationSemaphore Dependents
		ImageName ImageSaveTime ChangeFileName MaxInterruptLatency
		InterruptLatencyGoal VMSelectors DynamicCodeGCTrigger
		DynamicCodeLimit JustInTimeCompilationEnabled'
	poolDictionaries:''
	category:'System-Support'
!

Object subclass:#BinaryModuleDescriptor
	instanceVariableNames:'name type id dynamic classNames pathName libraryName timeStamp'
	classVariableNames:''
	poolDictionaries:''
	privateIn:ObjectMemory
!

!ObjectMemory class methodsFor:'documentation'!

caching
"
    The system uses various caches to speed up method-lookup.
    Currently, there is a three-level cache hierarchy:

        inline-cache            keeps the target of the last send at the caller-
                                side (i.e. every send goes through its private 
                                1-slot inline-cache, where the address of the last
                                called function at this call location is kept.)

        polymorph-inline-cache  keeps a limited list of all targets ever reached 
                                at this call location. The list is automatically 
                                flushed if it grows too large, or the overall number
                                of poly-chache entries exceeds a limit.

        method-lookup-cache     a global cache. Hashes on class-selector pairs,
                                returning the target method.

    Whenever methods are added or removed from the system, or the inheritance 
    hierarchy changes, some or all caches have to be flushed.
    The flushXXX methods perform the task of flushing various caches.
    All standard methods in Behavior call for cache flushing, when things change;
    however, if you use the low level access methods in Behavior
    (for example: #setSuperclass:) special care has to be taken.

    In some situations, not all caches need flushing, for example a change
    in an interpreted method (currently) needs no flushing of the inline caches.
    Also, flushing can be limited to entries for a specific class for most changes.

    To be 'on the brigth side of live', use ObjectMemory>>flushCaches (which
    flushes all of them), when in doubt of which caches should be flushed. 
    It is better flush too much - otherwise you may end up in a wrong method after 
    a send (you wont really see a difference in speed after the flush, anyway).
"
!

copyright
"
 COPYRIGHT (c) 1992 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class contains access methods to the system memory and the VM.

    In previous ST/X versions, this stuff used to be in the Smalltalk class.
    It has been separated for better overall class structure and modularisation.
    There are no instances of ObjectMemory - all is done in class methods.
    (this is a functional interface).

    Many methods here are for debuging purposes, for developers
    or experimental, and therefore not standard.
    Do not depend on them being there - some may vanish ...
    (especially those, that depend on a specific GC implementation)
    Most of the stuff found here is not available, or different or called
    different in other smalltalk implementations. Be aware, that using these
    interfaces (especially: depending on them) may make your application
    non portable.

    See more documentation in -> caching
                              -> interrupts
                              -> garbageCollection

    [Class variables:]

        InternalErrorHandler            gets informed (by VM), when some runtime
                                        error occurs (usually fatal)

        UserInterruptHandler            gets informed (by VM) when CNTL-C is pressed
        TimerInterruptHandler           gets alarm timer interrupts (from VM)
        SpyInterruptHandler             another alarm timer (from VM)
        StepInterruptHandler            gets single step interrupts (from VM)
        ExceptionInterruptHandler       gets floating point exceptions (from VM)
        ErrorInterruptHandler           gets primitive errors (from VM)
        MemoryInterruptHandler          gets soon-out-of-memory conditions (from VM)
        SignalInterruptHandler          gets unix signals (from VM)
        ChildSignalInterruptHandler     gets child death signals (from VM)
        DisposeInterruptHandler         gets informed, when an object is disposed from 
                                        a shadowArray (from VM)
        RecursionInterruptHandler       gets recursion limit violations (from VM)
        IOInterruptHandler              gets SIGIO unix signals (from VM)
        CustomInterruptHandler          gets custom interrupts (from VM)

        InterruptLatencyMonitor         if nonNil, that one will be notified (by the VM)
                                        with an interruptLatency:millis message for every 
                                        interrupt and gets the delay time
                                        (between the time when the signal arrived and
                                         when it was really delivered) as argument.
                                        This can be used to create a statistic for
                                        realtime systems.

        RegisteredErrorInterruptHandlers
                                        associates errorID (as passed from primitive
                                        to the __errorInterruptWithID() function)
                                        with handlers.

        IncrementalGCLimit              number of bytes, that must be allocated since
                                        last full garbage collect to turn the incremental
                                        collector on (at idle time).

        FreeSpaceGCLimit                low limit on freeSpace at which incremental
                                        gc starts to run at idle time.

        FreeSpaceGCAmount               amount to allocate once freeSpace drops
                                        below FreeSpaceGCLimit

	DynamicCodeGCTrigger		amount of generated dynamically compiled code 
				        to trigger the incremental garbage collector
					nil means: no trigger

	DynamicCodeLimit		max. amount of space allocated for dynamically compiled code
					nil means: unlimited.

        Dependents                      keep my dependents locally (its faster) for
                                        all those registries

        LowSpaceSemaphore               a semaphore signalled whenever the system is
                                        running in low memory (i.e. the memory manager
                                        ran into memory shortage and feels that it
                                        may soon no longer be able to grant allocation requests).
                                        You can have a process waiting on this semaphore
                                        which starts to remove (i.e. nil-out) objects
                                        or preform other cleanup actions.
                                        
        AllocationFailureSignal         signal raised when a new fails (see Behavior)
                                        When this signal is raised, the meomory manager
                                        is really in trouble (i.e. above feelings where
                                        correct)

        BackgroundCollectProcess        created by startBackgroundCollectorAt:

        BackgroundFinalizationProcess   created by startBackgroundFinalizationAt:

        ImageName                       name of the current image (or nil)

        ImageSaveTime   <AbsoluteTime>  timestamp when this image was saved


    Warning:
      The InterruptHandler variables are known by the runtime system -
      they are the objects that get an interrupt message when the event
      occurs. You may not remove any of them.

    [author:]
        Claus Gittinger
"
!

garbageCollection
"
    Currently, Smalltalk/X uses a two-level memory hierachy (actually, there
    are more memory regions used for stack, permanent objects, symbols etc.
    but for the following discussion, these are not of interest).

  newSpace:

    Objects are created in a so-called newSpace, which is relatively small.
    This newSpace is cleaned by a so called ``scavenge''-operation, whenever 
    becoming full. Scavenging means, that all still-live objects (i.e. referenced
    by some other) are copied over to another memory area, leaving all unreachable
    objects as garbage behind. Thus, the newSpace actually consists of two semispaces,
    of whih only one is active - the other being used only while objects are
    copied.
    After this copying, these two semispaces exchange their roles - i.e. reachable
    objects are copied ping-pong like between these semispaces.
    Once an object survives enough of these copying operations, the next scavenge 
    will move it into the so called oldSpace, which is much larger, and not
    processed by the scavenger. 
    This movement of an object from newSpace to oldSpace is called ``tenure''.
    (this avoids objects being copied around forever).
    Once tenured, an object is no longer contained in the newSpace, and
    thus ceases to create any scavenging overhead after that.

    Scavenging occurs automatically, and is usually done fast enough to go 
    unnoticed (typically, it takes some 5 to 40ms to perform a scavenge, 
    depending on how many live objects are in the newspace).

    Interestingly, the scavenger performs better, if many garbage objects
    are to be reclaimed, since less object-copying has to be done. Therefore,
    the best-case scavenge time is almost zero, if there is only garbage in
    the newSpace. In contrast, the worst-case is when all newSpace objects are still
    living. Thus, from a newSpace collectors viewPoint, it makes sense to get
    objects out of the way as fast as possible. However the oldSpace is
    collected much less frequently and the cost to reclaim an oldspace object
    is much higher (actually, the cost to reclaim a newspace object is zero -
    its the survival of objects which we have to pay for).
    Therefore, from an oldSpace collectors point of view, its preferable to
    keep objects in the newSpace as long as possible.

    To honor this conflicting situation, the system uses an adaptive tenure-count,
    which adjusts the number of scavenges required for tenure (the so called 
    'tenureAge') according to the fill-grade of the newSpace.
    If the newSpace is relatively empty, it tries to keep objects longer there.
    The controlling parameters of the tenure age can be changed dynamically,
    detailed information is available upon request.

    The exact speed of the scavenger depends mostly on the speed of your memory
    interface (and, since most of todays memories have access times in the order
    of 50-100ns, the raw CPU speed does not correlate linear with the GC speed).
    Measurements give roughly 40ms for a full 400k newSpace 
    (i.e. all objects survive) on a 486/50 - this only drops to some 20-30ms on a P5.

    The upper bounds of the scavenge blocking time can be controlled by changing
    the size of the newSpace - either via a command line argument, or even dynamically
    by Objectmemory>>newSpaceSize:. Smaller sizes lead to shorter blocking periods,
    but greater absolute GC overhead. The default (400k) seems to be a good compromise.
    (if you are not happy with it, try playing around with the settings)

  oldSpace:

    To reclaim oldspace, the system uses three algorithms: mark&sweep, a copying
    (and compressing) baker-type collector and an incremental mark&sweep.

    The mark&sweep runs whenever the oldspace becomes full, putting dead objects
    onto a free list. If a memory request cannot be served from this freelist,
    and the total size of objects on the freelist exceeds a threshold, the system
    will compress the oldspace to make the free-space into one big area.
    This compress is done by copying all live objects into a newly allocated
    area, and freeing the previous memory afterwards (baker collector).
    Since a compressing oldspace collect leads to a noticable pause of the system,
    the memory manager tries hard to avoid oldspace compression.
    (well, if enough real memory is available to hold both spaces in physical
     memory, the compress is actually pretty fast).

    The incremental mark&sweep runs in the background, whenever the system is idle
    (see ProcessorSceduler>>waitForEventOrTimeout), or alternatively as a low or high
    priority background process (see ObjectMemory>>startBackgroundCollector). 
    Like the normal mark&sweep, this incremental collector follows object references 
    and marks reachable objects on its way. However, this is done 'a few objects-at-a-time',
    to not disrupt the system noticably. Currently, there are some (theoretical) and in
    practice never occurring situations, in which the incremental GC still creates noticable
    delays. A current project is involved with this and a future version of ST/X (ST/X-RT)
    will be available which shows deterministic worst case behavior in its GC pauses 
    (this will be provided as an additional add-on option - certainly not for free ;-).

    Currently, incremental GC blockings are in the order of 10-70ms.
    There is one catch with low priority background IGC: if there is never any idle
    time available (i.e. all processes run all the time), it would never get a chance
    to do any collection work. To handle this case, a background IGC can also be started
    as a high priority process, which gives up the cpu (by delaying on the time) after
    every IGC step. A high priority background collector will always make progress
    and eventually finish a GC cycle. However, it may have more of an influence on 
    the other processes. So, its up to you, to decide ...


    Incremental garbage collection is controlled by the variables 
    'IncrementalGCLimit', 'FreeSpaceGCLimit' and 'FreeSpaceGCAmount':

      the ProcessorScheduler will perform incremental GC steps at idle time, 
      if the total space allocated since the last full collect exceeds 
      'IncrementalGCLimit',
      or if there are less than 'FreeSpaceGCLimit' bytes available in free store.
      If after the incrementalGC, less than 'FreeSpaceGCLimi't bytes are available,
      'FreeSpaceGCAmount' more bytes are requested from the memory manager.

    The defaults are set in ObjectMemory>>initialize and can be changed in your 
    startup 'smalltalk.rc'-file. Setting them to nil turns incremental GC off.


    For example, setting 'IncrementalGCLimit' to 500000 will start the background collector
    whenever 500k bytes have been allocated - usually very seldom. Setting it to some
    small number (say 10000) will have it run very often.

    Setting 'FreeSpaceGCAmount' to (say) 1meg lets the system try to always keep
    1meg of freeSpace. If less memory is available, more oldSpace will be allocated.
    Keeping some memory in the pocket may prevent the system from running into a blocking
    GC if memory is allocated in peaks (but only, if the incremental GC can keep up with 
    allocation rate). The trigger level 'FreeSpaceGCLimit' should be below that amount;
    to avoid excessive incremental GC activity (say 1/4 if the amount).

    Having the background GC running often should not hurt the performance of your 
    smalltalk processes, since the IGC only runs at times when no ST processes are runnable. 
    (there are some short delays in event processing, since the IGC's steps may take 
    some XX ms.) 
    However, if you are not alone on your machine (i.e. a timesharing system) or 
    you have other Unix processes to run, you should not run the IGC too often, 
    since it may hurt OTHER users/unix processes.

    Since this collector only runs at idle times, even a low priority background 
    process will prevent it from doing its work. You may want to start a somewhat
    higher priority background collect (say at prio 4), which also preempts these
    background processes. (see ObjectMemory>>startBackgroundCollectorAt:).

    Beginning with 2.10.4, a third space, called symSpace has been added.
    Objects in this space are never moved or garbage collected.
    This space is used for (some) symbols only.

    Beginning with 2.10.5, a fourth space, called fixSpace has been added.
    Objects in this space are never moved or garbage collected.
    This space is used for constant objects (true, false, some basic classes etc.).

    A plan for 2.11 is to offer an arbitrary number of spaces, which can be
    attached and detached at runtime. This will allow easy share of object
    with remote systems and separating objects into a per application/package
    space. (be prepared for changes in the future and make your application
    independ of the VM internals)

  default setup:

    The following table lists some default settings and means for changing them:

            what        default     change by           
                                    command line arg    dynamically
    -----------------------------------------------------------------------
        newSpace size     400k      -Mnew nnn           newSpaceSize:nnn

        oldSpace size    3000k      -Mold nnn           moreOldSpace:
                                                        announceSpaceNeed:
                                                        collectGarbage

        max tenure age     29                           lockTenure:
                                                        avoidTenure:
                                                        (sets it to infinity)
        
        adaptive tenure                                 tenureParameters

        oldSpace
        compressor      enabled     -Msingle            -

        limit for
        old-compress     8000k      -                   oldSpaceCompressLimit:

        chunk size
        to increase
        oldpPace          256k      -                   oldSpaceIncrement:
        
        prefer moreOld
        to doing GC      false      -                   fastMoreOldSpaceAllocation:

        limit for
        above                -      -                   fastMoreOldSpaceLimit:

        keep size for        -      -                   freeSpaceGCAmount:
        IGC

        low water
        trigger for IGC      -      -                   freeSpaceGCLimit:

        allocated
        trigger for IGC   500k      -                   incrementalGCLimit

        
    By default, no incremental GC is started by the system; however,
    the standard startup script starts a low prio background incremental GC process.
    You have to edit your startup files to change this. 
    A suggested configuration (used by the author and the default) is:

        ' keep 1meg in the pocket '

        ObjectMemory freeSpaceGCAmount:1000000. 

        ' start incrementalGC when freespace drops below 250k '
        ' or 500k of oldSpace has been allocated              '

        ObjectMemory freeSpaceGCLimit:250000.                 '
        ObjectMemory incrementalGCLimit:500000.               '

        ' collect as a background process (the default is: at idle times)
        ' this means that running cubes or other demo processes are suspended
        ' for the collect; change the prio to below 4 if you want them to continue

        ObjectMemory startBackgroundCollectorAt:5.            '
        ObjectMemory startBackgroundFinalizationAt:5.         '

        ' quickly allocate more space (i.e. avoid blocking collects)
        ' up to 8meg - then start to collect if more memory is needed.

        ObjectMemory fastMoreOldSpaceLimit:8*1024*1024.       '
        ObjectMemory fastMoreOldSpaceAllocation:true.         '

  hints & tricks:

    normally, there is no need to call for an explicit garbage collection, or
    modify the default parameters.
    The memory system should adapt reasonable and provide good performance 
    for a wide range of allocation patterns (see Example3 below for an exception).

    However, there may be situations, in which hints and/or explicit
    control over allocation can speedup your programs; but please:

      - if you think you have to play around with the memory policies,
        first check your program - you may find useless allocations
        or bad uses of collections. A typical error that is made is to
        create large collections using the #, (comma) concatenation method,
        which shows square behavior, since it allocates many, many temporary
        collections. Also, watch out for #copyWith:, #add: etc.
        All of these create a new collection. Remember, that most collections
        offer methods to preallocate some space; for example, 'Set new:' creates
        an empty set, but preallocates space to avoid resizing over and over.

        An especially bad performace dog is to use #add: on fix-size collection
        objects (such as Strings or Arrays), since in addition to allocating
        lots of garbage, a #become: operation is required for EACH element
        added. NEVER use Arrays for growing/shrinking data - use OrderedCollection
        instead. (if you really need an array, use asArray afterwards)

      - if you are going to allocate huge data structures, think about
        optimizing space. For example, if you allocate a million instances of
        some object, each added instance variable makes up 4Mb of additional 
        memory need.
        Also, for Byte-valued, Integer-valued and Float like objects, special
        collections are provided, which store their values directly inside (instead
        of a reference to the object). A FloatArray consisting of 1 million floats
        requires about 4mb of memory, while an Array of Floats requires 4mb for the
        references to the floats, PLUS 20Mb for the floats themself.

      - check if you really need fast access to all of these objects; you may
        try to only keep some subset in memory, and use binary storage or
        (if this is too slow) optimized store/retrieve methods and keep the bigger
        part in a file. 
        (How about a DiskArray class, which does this transparently ?
         See the FileText class for some ideas and something to start with ...)


    Hint / Example 1: 
      you are about to allocate a huge data structure, which is known to
      survive long. In this case, it is better to have these objects move into the
      oldspace sooner, to avoid the copying overhead during scavenges.

      To do this, you can call ObjectMemory>>tenure after allocation, which
      forces all new-objects immediately into the oldspace. 
      Make certain, that not to many (ideally no) short-living objects are in the
      newspace when doing this.

      Another alternative is to tell the system that all allocation should be
      done directly in the oldspace. This completely avoids the scavenging overhead
      for these objects. To do so, use ObjectMemory>>turnGarbageCollectorOff
      before the allocation, and ObjectMemory>>turnGarbageCollectorOn afterwards.
      Keep in mind, that do-loops may allocate block-objects and other temporaries,
      so there is a danger of making things worse due to having all those temporaries
      in the oldspace afterwards. (which is not a fatal situation, but will
      force the system to do an oldspace collect earlier, which may not be your
      intention).


   Hint / Example 2:
      you know in advance, that a certain (big) amount of memory will be needed.
      For example, the fileBrowser wants to show a huge file in its text-view.
      In this case, it is better to tell the memory system in advance, how much
      memory will be needed, since otherwise many compresses and reallocations will
      occur (the memory system will allocate additional memory in chunks of smaller
      256k pieces, if a compress failes. Thus, if you are going to allocate (say) 1Mb of 
      strings, it will perform 5 compressing GC's).

      This is done using ObjectMemory>>moreOldSpace: or ObjectMemory announceSpaceNeed:.
      In the above example, you would do 'ObjectMemory announceSpaceNeed:500000', which 
      avoids those annoying 5 compressing GC's.
      BTW: if you have other smalltalk processes (threads) running which should not be
      paused if possible, it is better to use #announceSpaceNeed. This tries to avoid 
      pausing in other processes and sometimes succeeds, while moreOldSpace will always 
      block the whole system for a while. However, there is no 'no-pause' guarantee.

      The amount of automatic increase (in case the oldSpace becomes full) is 256k by
      default. This number can be changed with ObjectMemory>>oldSpaceIncrement:.


    Hint / Example3:
      There are rare cases, when an explicit GC makes a difference: since
      object finalization is done at GC time, objects which keep operatingSystem
      resources may be finalized late. This is normally no problem, except if
      the system is running out of resources. For example, allocating new colors
      may fail if many colors have already been allocated in the past - even
      though these colors are actually free. The Depth8Image calls for an
      explicit GC, whenever it fails to allocate a color for a bitmap, to force
      finalization of free, but not yet finalized colors.
      The same is true for things like file descriptors, if fileStreams are not closed
      correctly.

    Hint 4:
      If you run in too small of physical memory, the incremental GC may have a
      bad effect on your working set: since it touches pages (which may otherwise
      not be needed at the moment, the operating system is forced to steal other
      (possibly more useful) pages from your set of incore pages.
      Although ST/X would be willing to fix this behavior (by telling the OS about
      its page requirements, many OS's do not listen. The SGI-Iris, For example ignores
      the madvice system calls - other systems do not implement it.
      The trouble is that the standard LRU paging strategy is exactly the worst for
      a program which sequentially scans its memory once in a single direction ...
      You may get better performance, if you turn off the incremental GC while
      processing a big data structure.


    Warning: many of the methods found here are not standard and may not even be available in
    future versions of ST/X. Use them only in very special situations or experiments.

    Let me know about additional special features you think are useful, and about
    special features you are using - this provides the feedback required to decide
    which methods are to be removed, kept or enhanced in future versions.
"
!

interrupts
"
    Handling of interrupts (i.e. unix-signals) is done via handler objects, which
    get a #XXXInterrupt-message sent. This is more flexible than (say) signalling
    a semaphore, since the handler-object may do anything to react on the signal
    (of course, it can also signal a semaphore to emulate the above behavior).

    Another reason for having handler objects is that they allow interrupt handling
    without any context switch, for high speed interrupt response.
    However, if you do this, special care is needed, since it is not defined, 
    which (smalltalk-)process gets the interrupt and will do the processing 
    (therefore, the default setup installs handlers which simply signal a semaphore 
     and continue the current process).

    Typically, the handlers are set during early initialization of the system
    by sending 'ObjectMemory XXXInterruptHandler:aHandler' and not changed later.
    (see Smalltalk>>initialize or ProcessorScheduler>>initialize).
    To setup your own handler, create some object which responds to #xxxInterrupt,
    and make it the handler using the above method.

    Interrupt messages sent to handlers are:
        internalError:<someString>      - internal interpreter/GC errors
        userInterrupt                   - ^C interrupt
        customInterrupt                 - custom interrupt
        ioInterrupt                     - SIGIO interrupt
        timerInterrupt                  - alarm timer (SIGALRM)
        errorInterrupt:<id>             - errors from other primitives/subsystems
                                          (DisplayError)
        spyInterrupt                    - spy timer interrupt (SIGVTALARM)
        stepInterrupt                   - single step interrupt
        disposeInterrupt                - finalization required
        recursionInterrupt              - recursion (stack) overflow
        memoryInterrupt                 - soon running out of memory
        fpExceptionInterrupt            - floating point exception (SIGFPE)
        childSignalInterrupt            - death of a child process (SIGCHILD)
        signalInterrupt:<number>        - unix signal (if other than above signals)

    To avoid frustration in case of badly set handlers, these messages
    are also implemented in the Object class - thus anything can be defined
    as interrupt handler. However, the VM will not send any
    interrupt message, if the corresonding handler object is nil
    (which means that nil is a bad choice, if you are interested in the event).

    Interrupt processing is not done immediately after the event arrives: 
    there are certain ``save-places'' at which this handling is performed
    (message send, method return and loop-heads).
    If not explicitely enabled, primitive code is never interrupted.
    However, if you do enable interrupts in your primitive (see ExternalStream as example),
    be prepared for your objects to move around ... therefore, these have to
    be coded very carefully.

    Interrupts may be disabled completely (OperatingSystem blockInterrupts) and 
    reenabled (unblockInterrupts) to allow for critical data to be manipulated.
    The above are low-level primitive entries - you better use #valueUninterruptably,
    which cares for unwinds, long returns and restores the blocking state.

    Every process has its own interrupt-enable state which is switched
    when processes switch control (i.e. you cannot block interrupts across
    a suspend, delay etc.). 
    However, the state will be restored after a resume.
"
! !

!ObjectMemory class methodsFor:'initialization'!

initialize
    "initialize the class"

    AllocationFailureSignal isNil ifTrue:[
        AllocationFailureSignal := ErrorSignal newSignalMayProceed:true.
        AllocationFailureSignal nameClass:self message:#allocationFailureSignal.
        AllocationFailureSignal notifierString:'allocation failure'.

        MallocFailureSignal := AllocationFailureSignal newSignalMayProceed:true.
        MallocFailureSignal nameClass:self message:#mallocFailureSignal.
        MallocFailureSignal notifierString:'(malloc) allocation failure'.

        LowSpaceSemaphore := Semaphore new name:'LowSpaceSemaphore'
    ].
    DisposeInterruptHandler := self.
    IncrementalGCLimit := 500000.
    FreeSpaceGCLimit := FreeSpaceGCAmount := nil.
    MemoryInterruptHandler := self.
    ExceptionInterruptHandler := self.

    VMSelectors := #( #noByteCode #invalidCodeObject #invalidByteCode #invalidInstruction
                      #tooManyArguments #badLiteralTable #receiverNotBoolean: #typeCheckError
                      #integerCheckError #wrongNumberOfArguments: #privateMethodCalled
                      #doesNotUnderstand: #invalidReturn: #invalidReturnOrRestart:
                      #userInterrupt #internalError: #spyInterrupt #timerInterrupt #stepInterrupt
                      #errorInterrupt:with: #disposeInterrupt #recursionInterrupt
                      #memoryInterrupt #fpExceptionInterrupt #signalInterrupt: #childSignalInterrupt
                      #ioInterrupt #customInterrupt #schedulerInterrupt #contextInterrupt
                      #interruptLatency:receiver:class:selector:vmActivity:id:).

    "Modified: 7.11.1996 / 19:53:42 / cg"
! !

!ObjectMemory class methodsFor:'ST-80 compatibility'!

availableFreeBytes
    "return the amount of free memory
     (both in the compact free area and in the free lists)"

    ^ self freeSpace + self freeListSpace

    "
     ObjectMemory availableFreeBytes 
    "

    "Modified: 29.1.1997 / 23:43:05 / cg"
!

bytesPerOOP
    "return the number of bytes an object reference (for example: an instvar)
     takes"

%{  /* NOCONTEXT */
    RETURN(__MKSMALLINT(sizeof(OBJ)));
%}

    "
     ObjectMemory bytesPerOOP 
    "
!

bytesPerOTE
    "return the number of overhead bytes of an object.
     i.e. the number of bytes in every objects header."

%{  /* NOCONTEXT */
    RETURN(__MKSMALLINT(OHDR_SIZE));
%}

    "
     ObjectMemory bytesPerOTE 
    "
!

compactingGC
    "perform a compacting garbage collect"

    self garbageCollect

    "Modified: 29.1.1997 / 23:43:25 / cg"
!

current
    "the 'current' ObjectMemory - thats myself"

    ^ self

    "Modified: 29.1.1997 / 23:43:39 / cg"
!

globalCompactingGC
    "perform a compacting garbage collect"

    self garbageCollect

    "Modified: 29.1.1997 / 23:43:47 / cg"
!

growMemoryBy:numberOfBytes
    "allocate more memory"

    ^ self moreOldSpace:numberOfBytes

    "Modified: 29.1.1997 / 23:44:01 / cg"
!

numOopsNumBytes
    "return an array filled with the number of objects
     and the number of used bytes."

    ^ Array with:(self numberOfObjects)
            with:(self bytesUsed)

    "
     ObjectMemory numOopsNumBytes 
    "

    "Modified: 29.1.1997 / 23:44:24 / cg"
!

versionId
    "return this systems version.
     For ST80 compatibility."

    ^ Smalltalk versionString

    "Created: / 27.10.1997 / 13:51:59 / cg"
    "Modified: / 28.10.1997 / 20:06:14 / cg"
! !

!ObjectMemory class methodsFor:'Signal constants'!

allocationFailureSignal
    "return the signal raised when an object allocation failed"

    ^ AllocationFailureSignal
!

mallocFailureSignal
    "return the signal raised when malloc memory allocation failed.
     (usually, this kind of memory is used with I/O buffers or other temporary
      non-Object storage)"

    ^ MallocFailureSignal
! !

!ObjectMemory class methodsFor:'VM messages'!

debugPrinting
    "return true, if various debug printouts in the VM
     are turned on, false of off."

%{  /* NOCONTEXT */
    extern int __getDebugPrinting();

    RETURN (__getDebugPrinting() ? true : false);
%}
!

debugPrinting:aBoolean
    "turn on/off various debug printouts in the VM
     in case of an error. For example, a double-notUnderstood
     leads to a VM context dump if debugPrinting is on.
     If off, those messages are suppressed.
     The default is on, since these messages are only printed for
     severe errors."

%{  /* NOCONTEXT */
    __setDebugPrinting( (aBoolean == true) );
%}
!

infoPrinting
    "return true, if various informational printouts in the VM
     are turned on, false of off."

%{  /* NOCONTEXT */
    extern int __getInfoPrinting();

    RETURN (__getInfoPrinting() ? true : false);
%}
!

infoPrinting:aBoolean
    "turn on/off various informational printouts in the VM.
     For example, the GC activity messages are controlled by
     this flags setting.
     The default is true, since (currently) those messages
     are useful for ST/X developers."

%{  /* NOCONTEXT */
    __setInfoPrinting( (aBoolean == true) );
%}
! !

!ObjectMemory class methodsFor:'access debugging'!

setTrapOnAccessFor:anObject
    "install an access trap for anObject;
     An accessSignal will be raised, whenever any instvar of anObject is either read or written.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 2) ? true : false);
%}
!

setTrapOnReadFor:anObject
    "install a read trap for anObject;
     An accessSignal will be raised, whenever any access into anObject occurrs.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 0) ? true : false);
%}
!

setTrapOnWriteFor:anObject
    "install a write trap for anObject;
     An accessSignal will be raised, whenever any instvar of anObject is written to.
     This is not supported on all architectures, therefore the return value
     (true of trap was installed ok, false if failed) should be checked."

%{  /* NOCONTEXT */
    RETURN (__addTrapOnAccess(anObject, 1) ? true : false);
%}
!

unsetAllTraps
    "remove all access traps"

%{  /* NOCONTEXT */
    __removeAllAccessTraps();
%} 
!

unsetTrapFor:anObject
    "remove any access trap for anObject."

%{  /* NOCONTEXT */
    __removeTrapOnAccess(anObject);
%}
! !

!ObjectMemory class methodsFor:'cache management'!

flushCaches
    "flush method and inline caches for all classes"

%{  /* NOCONTEXT */
    __flushMethodCache();
    __flushAllInlineCaches();
%}
!

flushCachesFor:aClass
    "flush method and inline caches for aClass"

%{  /* NOCONTEXT */
    __flushMethodCacheFor(aClass);
    __flushInlineCachesFor(aClass);
%}
!

flushCachesForSelector:aSelector
    "flush method and inline caches for aSelector"

%{  /* NOCONTEXT */
    __flushMethodCacheForSelector(aSelector);
    __flushInlineCachesForSelector(aSelector);
%}
!

flushInlineCaches
    "flush all inlinecaches"

%{  /* NOCONTEXT */
    __flushAllInlineCaches();
%}
!

flushInlineCachesFor:aClass withArgs:nargs
    "flush inlinecaches for calls to aClass with nargs arguments"

%{  /* NOCONTEXT */
    __flushInlineCachesForAndNargs(aClass, __intVal(nargs));
%}
!

flushInlineCachesForClass:aClass
    "flush inlinecaches for calls to aClass."

%{  /* NOCONTEXT */
    __flushInlineCachesFor(aClass);
%}
!

flushInlineCachesForSelector:aSelector
    "flush inlinecaches for sends of aSelector"

%{  /* NOCONTEXT */
    __flushInlineCachesForSelector(aSelector);
%}
!

flushInlineCachesWithArgs:nargs
    "flush inlinecaches for calls with nargs arguments"

%{  /* NOCONTEXT */
    __flushInlineCaches(__intVal(nargs));
%}
!

flushMethodCache
    "flush the method cache"

%{  /* NOCONTEXT */
    __flushMethodCache();
%}
!

flushMethodCacheFor:aClass
    "flush the method cache for sends to aClass"

%{  /* NOCONTEXT */
    __flushMethodCacheFor(aClass);
%}
!

flushMethodCacheForSelector:aSelector
    "flush the method cache for sends of aSelector"

%{  /* NOCONTEXT */
    __flushMethodCacheForSelector(aSelector);
%}
!

incrementSnapshotID
    "obsolete - do not use"

%{  /* NOCONTEXT */
    __bumpSnapshotID();
%}
!

snapshotID
    "return the internal snapshotID number.
     This is incremented when an image is restarted, and
     stored with the image.
     Not for normal users, this is used by the VM to invalidate
     caches which are stored with the image"

%{  /* NOCONTEXT */
    RETURN ( __MKSMALLINT( __snapshotID() ));
%}
    "
     ObjectMemory snapshotID     
    "
!

trapRestrictedMethods:trap
    "Allow/Deny execution of restricted Methods (see Method>>>restricted:)

     Notice: method restriction is a nonstandard feature, not supported
     by other smalltalk implementations and not specified in the ANSI spec.
     This is EXPERIMENTAL - and being evaluated for usability.
     It may change or even vanish (if it shows to be not useful)."

    |oldTrap|

%{ 
	if (__setTrapRestrictedMethods(trap == true))
	    oldTrap = true;
	else
	    oldTrap = false;
%}.

    (trap and:[oldTrap not]) ifTrue:[
	self flushCaches
    ].
    ^ oldTrap

    "
	ObjectMemory trapRestrictedMethods:true
	ObjectMemory trapRestrictedMethods:false
    "
! !

!ObjectMemory class methodsFor:'debug queries'!

addressOf:anObject
    "return the core address of anObject as an integer
     - since objects may move around, the returned value is invalid after the
     next scavenge/collect.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( nil );
    }
    if ((unsigned INT)anObject <= _MAX_INT) {
	RETURN ( __MKSMALLINT((INT)anObject) );
    }
    RETURN ( __MKUINT((INT)anObject) );
%}
    "
    |p|
    p := Point new.
    ((ObjectMemory addressOf:p) printStringRadix:16) printNL.
    ObjectMemory scavenge.
    ((ObjectMemory addressOf:p) printStringRadix:16) printNL.
    "
!

ageOf:anObject
    "return the number of scavenges, an object has survived
     in new space. 
     For old objects and living contexts, the returned number is invalid.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( 0 );
    }
    RETURN ( __MKSMALLINT( _GET_AGE(anObject) ) );
%}
    "
    |p|
    p := Point new.
    (ObjectMemory ageOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    ObjectMemory tenuringScavenge.
    (ObjectMemory spaceOf:p) printNL.
    "
!

dumpObject:someObject
    "low level dump an object.
     WARNING: this method is for ST/X debugging only 
	      it may be removed (or replaced by a noop) without notice"

%{
    __dumpObject__(someObject);
%}
    "
     ObjectMemory dumpObject:true
     ObjectMemory dumpObject:(Array new:10) 
     ObjectMemory dumpObject:(10@20 corner:30@40) 
    "
!

flagsOf:anObject
    "For debugging only.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( nil );
    }
    RETURN ( __MKSMALLINT( anObject->o_flags ) );
%}
    "
F_ISREMEMBERED  1       /* a new-space thing being refd by some oldSpace thing */
F_ISFORWARDED   2       /* a forwarded object (you will never see this here) */
F_DEREFERENCED  4       /* a collection after grow (not currently used) */
F_ISONLIFOLIST  8       /* a non-lifo-context-referencing-obj already on list */
F_MARK          16      /* mark bit for background collector */
    "

    "
     |arr|

     arr := Array new.
     arr at:1 put:([thisContext] value).
     (ObjectMemory flagsOf:anObject) printNL
    "
!

globalRefChainTo:anObject
     |levels objects setOfOwners|

     levels := OrderedCollection new.
     objects := WeakIdentitySet with:anObject.

     [objects includes:Smalltalk] whileFalse:[
         setOfOwners := WeakIdentitySet new.

        objects do:[:o |
           |owners|

           owners := o allOwners.
           owners notNil ifTrue:[
               owners := owners select:[:el | (el isMemberOf:WeakArray) not].
               setOfOwners addAll:owners.
           ].

"/           (no instVarNamed:#keyArray) atAllPut:nil.
"/           no := nil.
"/           (owners instVarNamed:#keyArray) atAllPut:nil.
"/           owners := nil.
        ].

        levels add:objects.
'done with level: ' print. levels size print. ' found ' print. objects size print. ' refs' printCR.

        levels do:[:alreadyFound |
            setOfOwners remove:(alreadyFound instVarNamed:#keyArray) ifAbsent:nil.
            alreadyFound do:[:o |
                setOfOwners remove:o ifAbsent:nil
            ]
        ].

        setOfOwners remove:(objects instVarNamed:#keyArray) ifAbsent:nil.
"/        setOfOwners printNL.

        objects := setOfOwners.
        objects isEmpty ifTrue:[
"/            'none found' printNL.
            ^ nil
        ].
     ].

"/     '' printNL.

     ^ self 
        refChainFrom:Smalltalk
        to:anObject 
        inRefSets:levels reverse
        startingAt:1 

     "
      self globalRefChainTo:Point new
     "
     "
      Smalltalk at:#foo put:Point new.
      self globalRefChainTo:(Smalltalk at:#foo)
     "
     "
      |p|
      p := Point new.
      Smalltalk at:#foo put:(#x -> p).
      self globalRefChainTo:p
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> a).
      self globalRefChainTo:a 
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> (Array with:a)).
      self globalRefChainTo:a 
     "
     "
      |a|
      a := Array new:1. a at:1 put:a.
      Smalltalk at:#foo put:(#x -> (Array with:a with:(#y -> a))).
      self globalRefChainTo:a 
     "
     "
      |p|
      p := Point new.
      Smalltalk at:#foo put:(WeakArray with:(#x -> (#y -> p))).
      self globalRefChainTo:p
     "
     "
      self globalRefChainTo:Transcript topView
     "

    "Modified: 20.5.1996 / 10:28:58 / cg"
!

objectAt:anAddress
    "return whatever anAddress points to as object.
     BIG BIG DANGER ALERT: 
	this method is only to be used for debugging ST/X itself 
	- you can easily (and badly) crash the system.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */
    int addr = __longIntVal(anAddress);

    if (addr) {
	RETURN ((OBJ)(addr));
    }
    RETURN (nil);
%}
!

printReferences:anObject
    "for debugging: print referents to anObject.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice
	use ObjectMemory>>whoReferences: or anObject>>allOwners."

%{
    __printRefChain(__context, anObject);
%}
!

refChainFrom:start to:anObject inRefSets:levels startingAt:index
     |chain names oClass|

     index > levels size ifTrue:[
        ^ OrderedCollection with:start
     ].

"/'set: ' print. (levels at:index) displayString printNL.
"/'start: ' print. start printNL.
     (levels at:index) do:[:el |
"/'trying: ' print. el printNL.
        (start references:el) ifTrue:[
"/'yes' printNL.
            chain := self 
                        refChainFrom:el 
                        to:anObject 
                        inRefSets:levels 
                        startingAt:(index+1).

            chain notNil ifTrue:[
                start == Smalltalk ifTrue:[
                    start keysAndValuesDo:[:key :val |
                        (val == el) ifTrue:[
                            chain addFirst:'Smalltalk:' , key.
                            ^ chain
                        ]
                    ]
                ] ifFalse:[
                    names := start class allInstVarNames.
                    oClass := start class.
                    1 to:oClass instSize do:[:i |
                        ((start instVarAt:i) == el) ifTrue:[
                            chain addFirst:(start -> ('[' , (names at:i) , ']')).
                            ^ chain
                        ].
                    ].
                    oClass isVariable ifTrue:[
                        oClass isPointers ifTrue:[
                            1 to:start basicSize do:[:i |
                                ((start basicAt:i) == el) ifTrue:[
                                    chain addFirst:(start -> ('[' , i printString , ']')).
                                    ^ chain
                                ]
                            ]
                        ]
                    ].
                ].

                chain addFirst:start.

                ^ chain
            ]
        ].
     ].
     ^ nil


     "
      |o a1 a2|

      o := Object new.
      a1 := Array with:o with:nil with:nil.
      a2 := Array with:a1 with:nil with:nil.
      a2 references:a1
     "

     "
      |o a1 a2 a3 a4 levels|

      o := Object new.
      a1 := Array with:o.
      a2 := Array with:a1.
      a3 := Array with:a2.
      a4 := Array with:a3.
      levels := Array with:(Array with:a3)
                      with:(Array with:a2)
                      with:(Array with:a1).

      self refChainFrom:a4 to:o inRefSets:levels startingAt:1. 
     "

    "Modified: 19.3.1996 / 23:22:41 / cg"
!

sizeOf:anObject
    "return the size of anObject in bytes.
     (this is not the same as 'anObject size').
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    RETURN ( __isNonNilObject(anObject) ? __MKSMALLINT(__qSize(anObject)) : __MKSMALLINT(0) )
%}
    "
     |hist big nw|

     hist := Array new:100 withAll:0.
     big := 0.
     ObjectMemory allObjectsDo:[:o |
	 nw := (ObjectMemory sizeOf:o) // 4 + 1.
	 nw > 100 ifTrue:[
	    big := big + 1
	 ] ifFalse:[
	    hist at:nw put:(hist at:nw) + 1
	 ].
     ].
     hist printNL.
     big printNL
    "
!

spaceOf:anObject
    "return the memory space, in which anObject is.
     - since objects may move between spaces, 
       the returned value may be invalid after the next scavenge/collect.
     WARNING: this method is for ST/X debugging only 
	      it will be removed without notice"

%{  /* NOCONTEXT */

    if (! __isNonNilObject(anObject)) {
	RETURN ( nil );
    }
    RETURN ( __MKSMALLINT( __qSpace(anObject) ) );
%}
! !

!ObjectMemory class methodsFor:'dependents access'!

dependents
    "return the collection of my dependents"

    ^ Dependents ? #()

    "Modified: / 26.1.1998 / 11:17:59 / cg"
!

dependents:aCollection
    "set the dependents collection"

    Dependents := aCollection
!

dependentsDo:aBlock
    "evaluate aBlock for all of my dependents.
     Since this is performed at startup time (under the scheduler),
     this is redefined here to catch abort signals.
     Thus, if any error occurs in a #returnFromSnapshot,
     the user can press abort to continue."

    |deps|

    deps := Dependents.
    deps notNil ifTrue:[
	deps do:[:each |
	    AbortSignal handle:[:ex |
		ex return       
	    ] do:[
		aBlock value:each
	    ]
	]
    ]
! !

!ObjectMemory class methodsFor:'enumerating'!

allInstancesOf:aClass do:aBlock
    "evaluate the argument, aBlock for all instances of aClass in the system.
     There is one caveat: if a compressing oldSpace collect
     occurs while looping over the objects, the loop cannot be
     continued (for some internal reasons). In this case, false
     is returned."

    |work|

%{  /* NOREGISTER - work may not be placed into a register here */
    __nonTenuringScavenge(__context);
    /*
     * allInstancesDo needs a temporary to hold newSpace objects
     */
    if (__allInstancesOfDo(&aClass, &aBlock, &work) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

allObjectsDo:aBlock
    "evaluate the argument, aBlock for all objects in the system.
     There is one caveat: if a compressing oldSpace collect
     occurs while looping over the objects, the loop cannot be
     continued (for some internal reasons). In this case, false
     is returned."

    |work|

%{  /* NOREGISTER - work may not be placed into a register here */
    __nonTenuringScavenge(__context);
    /*
     * allObjectsDo needs a temporary to hold newSpace objects
     */
    if (__allInstancesOfDo((OBJ *)0, &aBlock, &work) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

allOldObjectsDo:aBlock
    "evaluate the argument, aBlock for all old objects in the system.
     For debugging and tests only - do not use"
%{
    if (__allInstancesOfDo((OBJ *)0, &aBlock, (OBJ *)0) < 0) {
	RETURN (false);
    }
%}. 
    ^ true
! !

!ObjectMemory class methodsFor:'garbage collection'!

backgroundCollectorRunning
    "return true, if a backgroundCollector is running"

    ^ BackgroundCollectProcess notNil

    "
     ObjectMemory backgroundCollectorRunning
    "
!

compressingGarbageCollect
    "search for and free garbage in the oldSpace (newSpace is cleaned automatically) 
     performing a COMPRESSING garbage collect.
     This can take a long time - especially, if paging is involved
     (when no paging is involved, its faster than I thought :-).
     If no memory is available for the compress, or the system has been started with
     the -Msingle option, this does a non-COMPRESSING collect."
%{
    if (! __garbageCollect(__context)) {
	__markAndSweep(__context);
    }
%}

    "
     ObjectMemory compressingGarbageCollect
    "
!

garbageCollect
    "search for and free garbage in the oldSpace.
     This can take a long time - especially, if paging is involved."

    "/ used to be 
    "/    self compressingGarbageCollect 
    "/ here; changed to default to markAndSweep

    self markAndSweep

    "
     ObjectMemory garbageCollect
    "
!

gcStep
    "one incremental garbage collect step.
     Mark or sweep some small number of objects. This
     method will return after a reasonable (short) time.
     This is used by the ProcessorScheduler at idle times.
     Returns true, if an incremental GC cycle has finished."
%{
    extern int __incrGCstep();

    RETURN (__incrGCstep(__context) ? true : false);
%}
!

gcStepIfUseful
    "If either the IncrementalGCLimit or the FreeSpaceGCLimits have been
     reached, perform one incremental garbage collect step.
     Return true, if more gcSteps are required to finish the cycle,
     false if done with a gc round.
     If no limit has been reached yet, do nothing and return false.
     This is called by the ProcessorScheduler at idle times or by the 
     backgroundCollector."

    |done limit|

    AbortSignal handle:[:ex |
        "/ in case of abort (from the debugger),
        "/ disable gcSteps.
        done := true.
        IncrementalGCLimit := FreeSpaceGCLimit := nil.
        'ObjectMemory [error]: IGC aborted; turning off incremental GC' errorPrintCR
    ] do:[
        limit := IncrementalGCLimit.
        (limit notNil and:[self oldSpaceAllocatedSinceLastGC > limit]) ifTrue:[
            done := ObjectMemory gcStep
        ] ifFalse:[
            limit := FreeSpaceGCLimit.
            (limit notNil and:[(self freeSpace + self freeListSpace) < limit]) ifTrue:[
                done := ObjectMemory gcStep.
                done ifTrue:[
                    self moreOldSpaceIfUseful
                ].
            ] ifFalse:[
                limit := DynamicCodeGCTrigger.
                (limit notNil and:[self compiledCodeCounter > limit]) ifTrue:[
                    done := ObjectMemory gcStep.
                ] ifFalse:[
                    limit := DynamicCodeLimit.
                    (limit notNil and:[self compiledCodeSpaceUsed > limit]) ifTrue:[
                        done := ObjectMemory gcStep.
                    ] ifFalse:[
                        done := true
                    ]
                ]
            ]
        ].
    ].
    ^ done not

    "Modified: 10.1.1997 / 17:59:10 / cg"
!

incrementalGC
    "perform one round of incremental GC steps.
     The overall effect of this method is (almost) the same as calling 
     markAndSweep. However, #incrementalGC is interruptable while #markAndSweep
     is atomic and blocks for a while. The code here performs incremental
     GC steps, until one complete gc-cycle is completed. If running at a higher
     than userBackground priority, it will give up the CPU after every such
     step for a while.
     Thus this method can be called either from a low prio (background) process 
     or from a high prio process.
     (however, if you have nothing else to do, its better to call for markAndSweep,
      since it is faster)
     For example, someone allocating huge amounts of memory could
     ask for the possibility of a quick allocation using
     #checkForFastNew: and try a #incrementalGC if not. In many
     cases, this can avoid a pause (in the higher prio processes) due to 
     a blocking GC."

    |delay|

    Processor activeProcess priority > Processor userBackgroundPriority ifTrue:[
        delay := Delay forMilliseconds:1
    ].

    self gcStepIfUseful ifTrue:[
        [self gcStep] whileFalse:[
            delay notNil ifTrue:[delay wait]
        ]
    ].
    self moreOldSpaceIfUseful

    "
     ObjectMemory incrementalGC
     [ObjectMemory incrementalGC] forkAt:3
     [ObjectMemory incrementalGC] forkAt:9 
    "

    "Modified: / 12.11.1997 / 18:23:46 / cg"
!

markAndSweep
    "mark/sweep garbage collector.
     perform a full mark&sweep collect.
     Warning: this may take some time and it is NOT interruptable.
     If you want to do a collect from a background process, or have
     other things to do, better use #incrementalGC which is interruptable."
%{
    __markAndSweep(__context);
%}

    "
     ObjectMemory markAndSweep
    "
!

reclaimSymbols
    "reclaim unused symbols;
     Unused symbols are (currently) not reclaimed automatically,
     but only upon request with this method. 
     It takes some time to do this ... and it is NOT interruptable.
     Future versions may do this automatically, while garbage collecting."
%{
    __reclaimSymbols(__context);
%}
    "
     ObjectMemory reclaimSymbols
    "
!

scavenge
    "collect young objects, without aging (i.e. no tenure).
     Can be used to quickly get rid of shortly before allocated
     stuff. This is relatively fast (compared to oldspace collect).

     An example where a non-tenuring scavenge makes sense is when
     allocating some OperatingSystem resource (a Color, File or View) 
     and the OS runs out of resources. In this case, the scavenge may
     free some ST-objects and therefore (by signalling the WeakArrays
     or Registries) free the OS resources too.
     Of course, only recently allocated resources will be freed this
     way. If none was freed, a full collect will be needed."
%{
    __nonTenuringScavenge(__context);
%}

    "
     ObjectMemory scavenge
    "
!

startBackgroundCollectorAt:aPriority
    "start a process doing incremental GC in the background.
     Use this, if you have suspendable background processes which
     run all the time, and therefore would prevent the idle-collector
     from running. See documentation in this class for more details."

    |p|

    "/
    "/ its not useful, to run it more than once
    "/
    BackgroundCollectProcess notNil ifTrue:[
        BackgroundCollectProcess priority:aPriority.
        ^ self
    ].

    p :=
        [
            [
                |myDelay|

                myDelay := Delay forSeconds:5.

                [true] whileTrue:[
                    self gcStepIfUseful ifTrue:[
                        "
                         perform a full cycle
                        "
                        self incrementalGC
                    ].
                    "
                     wait a bit
                    "
                    myDelay wait
                ]
            ] valueOnUnwindDo:[
                BackgroundCollectProcess := nil
            ]
        ] newProcess.
    p name:'background collector'.
    p priority:aPriority.
    p restartable:true.
    p resume.
    BackgroundCollectProcess := p

    "
     the following lets the backgroundCollector run at prio 5
     whenever 100000 bytes have been allocated, OR freeSpace drops
     below 1meg. Having the system keep 1meg as reserve for peak allocation.

     Doing this may reduce pauses due to inevitable collects when running
     out of freeSpace, if the collector can keep up with allocation rate.
    "

    "
     ObjectMemory stopBackgroundCollector.

     ObjectMemory incrementalGCLimit:100000.
     ObjectMemory freeSpaceGCLimit:1000000.
     ObjectMemory startBackgroundCollectorAt:5.
    "

    "Modified: 24.1.1997 / 21:37:17 / cg"
!

stopBackgroundCollector
    "stop the background collector"

    BackgroundCollectProcess notNil ifTrue:[
	BackgroundCollectProcess terminate.
	BackgroundCollectProcess := nil
    ]

    "
     ObjectMemory stopBackgroundCollector
    "
!

tenure
    "force all living new stuff into old-space - effectively making
     all living young objects become old objects immediately.
     This is relatively fast (compared to oldspace collect).

     This method should only be used in very special situations:
     for example, when building up some long-living data structure
     in a time critical application.
     To do so, you have to do a scavenge followed by a tenure after the
     objects are created. Be careful, to not reference any other chunk-
     data when calling for a tenure (this will lead to lots of garbage in
     the oldspace).
     In normal situations, explicit tenures are not needed."
%{
    __tenure(__context);
%}

    "
     ObjectMemory tenure
    "
    "
     ... build up long living objects ...
     ObjectMemory scavenge.
     ObjectMemory tenure
     ... continue - objects created above are now in oldSpace ...
    "
!

tenuringScavenge
    "collect newspace stuff, with aging (i.e. objects old enough
     will be moved into the oldSpace).
     Use this for debugging and testing only - the system performs
     this automatically when the newspace fills up.
     This is relatively fast (compared to oldspace collect)"
%{
    __scavenge(__context);
%}

    "
     ObjectMemory tenuringScavenge
    "
!

verboseCompactingGC
    "ST80 compatibility; same as verboseGarbageCollect"

    self verboseGarbageCollect
!

verboseGarbageCollect
    "perform a compressing garbage collect and show some informational
     output on the Transcript"

    |nBytesBefore nReclaimed value unit|

    nBytesBefore := self oldSpaceUsed.
    self compressingGarbageCollect.
    nReclaimed := nBytesBefore - self oldSpaceUsed.
    nReclaimed > 0 ifTrue:[
        nReclaimed > 1024 ifTrue:[
            nReclaimed > (1024 * 1024) ifTrue:[
                value := nReclaimed // (1024 * 1024).
                unit := ' Mb.'
            ] ifFalse:[
                value := nReclaimed // 1024.
                unit := ' Kb.'
            ]
        ] ifFalse:[
            value := nReclaimed.
            unit := ' bytes.'
        ].
        Transcript show:'reclaimed '; show:value printString.
        Transcript showCR:unit
    ]

    "
     ObjectMemory verboseGarbageCollect
    "

    "Modified: 18.5.1996 / 15:44:13 / cg"
!

verboseGlobalCompactingGC
    "ST80 compatibility; same as verboseGarbageCollect"

    self verboseGarbageCollect
! !

!ObjectMemory class methodsFor:'garbage collector control'!

announceOldSpaceNeed:howMuch
    "announce to the memory system, that howMuch bytes of memory will be needed
     soon, which is going to live longer (whatever that means). 
     It first checks if the memory can be allocated without forcing a compressing
     GC. If not, the oldSpace is increased. This may also lead to a slow compressing
     collect. However, many smaller increases are avoided afterwards. Calling this
     method before allocating huge chunks of data may provide better overall performance.
     Notice: this is a nonstandard interface - use only in special situations."

    (self checkForFastNew:howMuch) ifFalse:[
	self incrementalGC.
	(self checkForFastNew:howMuch) ifFalse:[
	    self moreOldSpace:howMuch
	]
    ]

    "
     ObjectMemory announceOldSpaceNeed:1000000
    "
!

announceSpaceNeed:howMuch
    "announce to the memory system, that howMuch bytes of memory will be needed
     soon. The VM tries to prepare itself for this allocation to be performed
     with less overhead. For example, it could preallocate some memory in one
     big chunk (instead of doing many smaller reallocations later).
     Notice: this is a nonstandard interface - use only in special situations.
     Also, this does a background collect before the big chunk of memory is
     allocated, not locking other processes while doing so."

    (howMuch < (self newSpaceSize // 2)) ifTrue:[
	self scavenge.
    ].
    (self checkForFastNew:howMuch) ifFalse:[
	(howMuch > (self newSpaceSize // 2)) ifFalse:[
	    self scavenge.
	].
	self incrementalGC.
	(self checkForFastNew:howMuch) ifFalse:[
	    self moreOldSpace:howMuch
	]
    ]

    "
     ObjectMemory announceSpaceNeed:100000
    "
!

avoidTenure:flag
    "set/clear the avoidTenure flag. If set, aging of newSpace is turned off
     as long as the newSpace fill-grade stays below some magic high-water mark.
     If off (the default), aging is done as usual.
     If the flag is turned on, scavenge may be a bit slower, due to more
     objects being copied around. However, chances are high that in an idle
     or (almost idle) system, less objects are moved into oldSpace.
     Therefore, this helps to avoid oldSpace colelcts, in systems which go into
     some standby mode and are reactivated by some external event.
     (the avoid-flag should be turned off there, and set again once the idle loop
     is reentered).

    This is an EXPERIMENTAL interface."

%{  /* NOCONTEXT */
    __avoidTenure(flag == true ? 1 : 0);
%}
!

checkForFastNew:amount
    "this method returns true, if amount bytes could be allocated
     quickly (i.e. without forcing a full GC or compress).
     This can be used for smart background processes, which want to
     allocate big chunks of data without disturbing foreground processes
     too much. Such a process would check for fast-allocation, and perform
     incremental GC-steps if required. Thus, avoiding the long blocking pause 
     due to a forced (non-incremental) GC. 
     Especially: doing so will not block higher priority foreground processes.
     See an example use in Behavior>>niceBasicNew:.
     This is experimental and not guaranteed to be in future versions."

%{  /* NOCONTEXT */
    extern int __checkForFastNew();

    if (__isSmallInteger(amount)) {
	if (! __checkForFastNew(__intVal(amount))) {
	    RETURN (false);
	}
    }

%}.
    ^ true
!

dynamicCodeGCTrigger
    "return the dynamic code trigger limit for incremental GC activation.
     The system will start doing incremental background GC, whenever this amount
     of code was dynamically generated. 
     The default is nil; which disables this trigger"

    ^ DynamicCodeGCTrigger

    "
     ObjectMemory dynamicCodeGCTrigger
    "

    "Created: 16.10.1996 / 17:42:01 / cg"
!

dynamicCodeGCTrigger:numberOfBytesOrNil
    "set the dynamic code trigger limit for incremental GC activation.
     The system will start doing incremental background GC, whenever this amount
     of code was dynamically generated. 
     The default is nil; which disables this trigger"

    DynamicCodeGCTrigger := numberOfBytesOrNil

    "
     ObjectMemory dynamicCodeGCTrigger:50000
    "

    "Created: 16.10.1996 / 17:42:29 / cg"
!

dynamicCodeLimit
    "return the dynamic code limit.
     The system will start doing incremental background GC, whenever this amount
     of code has been generated (overall), and start to flush the code cache,
     if (after the GC), more code is still allocated. 
     The default is nil; which disables this trigger"

    DynamicCodeLimit := self getCompiledCodeLimit.
    ^ DynamicCodeLimit

    "
     ObjectMemory dynamicCodeLimit
    "

    "Created: 16.10.1996 / 17:43:37 / cg"
!

dynamicCodeLimit:nBytesOrNil
    "set the dynamic code limit.
     The system will start doing incremental background GC, whenever this amount
     of code has been generated (overall), and start to flush the code cache,
     if (after the GC), more code is still allocated. 
     The default is nil; which disables this trigger"

    DynamicCodeLimit := nBytesOrNil.
    self setCompiledCodeLimit:nBytesOrNil.

    "
     ObjectMemory dynamicCodeLimit:100000
    "

    "Created: 16.10.1996 / 17:43:58 / cg"
    "Modified: 16.10.1996 / 19:30:46 / cg"
!

fastMoreOldSpaceAllocation:aBoolean
    "this method turns on/off fastMoreOldSpace allocation.
     By default, this is turned off (false), which means that in case of
     a filled-up oldSpace, a GC is tried first before more oldSpace is allocated. 
     This strategy is ok for the normal operation of the system,
     but behaves badly, if the program allocates huge data structures (say a
     game tree of 30Mb in size) which survives and therefore will not be reclaimed
     by a GC.
     Of course while building this tree, and the memory becomes full, the system 
     would not know in advance, that the GC will not reclaim anything.

     Setting fastOldSpaceIncrement to true will avoid this, by forcing the
     memory system to allocate more memory right away, without doing a GC first.

     WARNING: make certain that this flag is turned off, after your huge data
     is allocated, since otherwise the system may continue to increase its
     virtual memory without ever checking for garbage.
     This method returns the previous value of the flag; typically this return
     value should be used to switch back."

%{   /* NOCONTEXT */
     extern int __fastMoreOldSpaceAllocation();

     RETURN (__fastMoreOldSpaceAllocation(aBoolean == true ? 1 : 0) ? true : false);
%}
    "
     |previousSetting|

     previousSetting := ObjectMemory fastMoreOldSpaceAllocation:true.
     [
	...
	allocate your huge data
	...
     ] valueNoOrOnUnwindDo:[
	ObjectMemory fastMoreOldSpaceAllocation:previousSetting
     ]
    "

    "
     |prev this|

     prev := ObjectMemory fastMoreOldSpaceAllocation:true.
     ObjectMemory fastMoreOldSpaceAllocation:prev.
     ^ prev
    "
!

fastMoreOldSpaceLimit:aNumber
    "this method sets and returns the fastMoreOldSpace limit.
     If fastMoreOldSpaceAllocation is true, and the current oldSpace size is
     below this limit, the memory manager will NOT do a GC when running out of
     oldSpace, but instead quickly go ahead increasing the size of the oldSpace.
     Setting the limit to 0 turns off any limit (i.e. it will continue to
     increase the oldSpace forwever - actually, until the OS refuses to give us
     more memory). The returned value is the previous setting of the limit."

%{   /* NOCONTEXT */
     extern int __fastMoreOldSpaceLimit();

     if (__isSmallInteger(aNumber)) {
	 RETURN ( __MKSMALLINT( __fastMoreOldSpaceLimit(__intVal(aNumber))));
     }
%}.
     self primitiveFailed
    
    "
     |prev this|

     prev := ObjectMemory fastMoreOldSpaceLimit:10*1024*1024.
     ObjectMemory fastMoreOldSpaceLimit:prev.
     ^ prev
    "
!

freeSpaceGCAmount
    "return the amount to be allocated if, after an incrementalGC,
     not at least FreeSpaceGCLimit bytes are available for allocation.
     The default is nil, which lets the system compute an abbpropriate value"

    ^ FreeSpaceGCAmount

    "
     ObjectMemory freeSpaceGCAmount
    "
!

freeSpaceGCAmount:aNumber
    "set the amount to be allocated if, after an incrementalGC,
     not at least FreeSpaceGCLimit bytes are available for allocation.
     The amount should be greater than the limit, otherwise the incremental
     GC may try over and over to get the memory (actually waisting time)."

    FreeSpaceGCAmount := aNumber

    "
     the following will try to always keep at least 1meg of free space
     (in the background) and start to do so, whenever the freeSpace drops
     below 250k.
    "
    "
     ObjectMemory freeSpaceGCLimit:250000.  
     ObjectMemory freeSpaceGCAmount:1000000.  
    "

    "
     turn it off (i.e. let the system  compute an appropriate amount ...)
    "
    "
     ObjectMemory freeSpaceGCAmount:nil.     
    "
!

freeSpaceGCLimit
    "return the freeSpace limit for incremental GC activation.
     The system will start doing incremental background GC, once less than this number 
     of bytes are available in the compact free space.
     The default is 100000; setting it to nil will turn this trigger off."

    ^ FreeSpaceGCLimit

    "
     ObjectMemory freeSpaceGCLimit
    "
!

freeSpaceGCLimit:aNumber
    "set the freeSpace limit for incremental GC activation.
     The system will start doing incremental background GC, once less than this number 
     of bytes are available for allocation.
     The default is nil; setting it to nil will turn this trigger off."

    FreeSpaceGCLimit := aNumber

    "
     the following will start the incrementalGC (in the background)
     whenever the freeSpace drops below 1meg of free space
    "
    "
     ObjectMemory freeSpaceGCLimit:1000000.  
    "

    "
     turn it off (i.e. let the system hit the wall ...)
    "
    "
     ObjectMemory freeSpaceGCLimit:nil.     
    "
!

incrementalGCLimit
    "return the  allocatedSinceLastGC limit for incremental GC activation.
     The system will start doing incremental background GC, once more than this number 
     of bytes have been allocated since the last GC. 
     The default is 500000; setting it to nil will turn this trigger off."

    ^ IncrementalGCLimit

    "
     ObjectMemory incrementalGCLimit
    "
!

incrementalGCLimit:aNumber
    "set the allocatedSinceLastGC limit for incremental GC activation.
     The system will start doing incremental background GC, once more than this number 
     of bytes have been allocated since the last GC. 
     The default is 500000; setting it to nil will turn this trigger off."

    IncrementalGCLimit := aNumber

    "
     ObjectMemory incrementalGCLimit:500000.  'do incr. GC very seldom'
     ObjectMemory incrementalGCLimit:100000.  'medium'
     ObjectMemory incrementalGCLimit:10000.   'do incr. GC very often'
     ObjectMemory incrementalGCLimit:nil.     'never'
    "
!

lockTenure:flag
    "set/clear the tenureLock. If the lock is set, the system
     completely turns off tenuring, and objects remain in newSpace (forever).
     Once this lock is set, the system operates only in the newSpace and no memory 
     allocations from oldSpace are allowed (except for explicit tenure calls). 
     If any allocation request cannot be resoved, the VM raises a memory interrupt,
     clears the lockTenure-flag and returns nil. Thus, it automatically falls back into
     the normal mode of operation, to avoid big trouble 
     (fail to allocate memory when handling the exception).

     This interface can be used in applications, which are guaranteed to have their
     working set completely in the newSpace AND want to limit the worst case
     pause times to the worst case scavenge time 
     (which itself is limitd by the size of the newSpace). 
     I.e. systems which go into some event loop after initial startup,
     may turn on the tenureLock to make certain that no oldSpace memory is 
     allocated in the future; thereby limiting any GC activity to newSpace scavenges only.

     This is an EXPERIMENTAL interface.
    "

%{
    if (flag == true) {
	__tenure(__context);
    }
    __lockTenure(flag == true ? 1 : 0);
%}
!

makeOld:anObject
    "move anObject into oldSpace.
     This method is for internal & debugging purposes only -
     it may vanish. Dont use it."
%{
    if (__moveToOldSpace(anObject, __context) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

maxOldSpace
    "return the maxOldSpace value. If non-zero, thats the limit for which the
     VM will try hard to not allocate more oldSpace memory. (its not a hard limit)
     If zero, it will allocate forever (until the OS wont hand out more).
     The default is zero."

%{  /* NOCONTEXT */
    extern unsigned __maxOldSpace();

    RETURN (__MKSMALLINT( __maxOldSpace(-1) )); 
%}
    "
     ObjectMemory maxOldSpace
    "
!

maxOldSpace:amount
    "set the maxOldSpace value. If non-zero, thats the limit for which the
     VM will try hard to not allocate more oldSpace memory. (its not a hard limit)
     If zero, it will allocate forever (until the OS wont hand out more).
     The default is zero.
     WARNING: 
        an oldSpace limit may lead to trashing due to exorbitant GC activity;
        its usually better to let it allocate more and page in/page out.
        Usually, the background GC will catch up sooner or later and reclaim
        the memory without blocking the system"

%{  /* NOCONTEXT */
    extern unsigned __maxOldSpace();

    if (__isSmallInteger(amount)) {
        RETURN (__MKSMALLINT( __maxOldSpace(__intVal(amount)) )); 
    }
%}
    "to change maximum to 4Megs:"
    "
     ObjectMemory maxOldSpace:4*1024*1024
    "
!

moreOldSpace:howMuch
    "allocate howMuch bytes more for old objects; return true if this worked,
     false if that failed.
     This is done automatically, when running out of space, but makes
     sense, if its known in advance that a lot of memory is needed to
     avoid multiple reallocations and compresses.
     On systems which do not support the mmap (or equivalent) system call,
     this (currently) implies a compressing garbage collect - so its slow.
     Notice: this is a nonstandard interface - use only in special situations."

%{  
    if (__isSmallInteger(howMuch)) {
	RETURN( __moreOldSpace(__context, __intVal(howMuch)) ? true : false );
    }
    RETURN (false);
%}
    "
     ObjectMemory moreOldSpace:1000000
    "
!

moreOldSpaceIfUseful
    "to be called after an incremental GC cycle;
     if freeSpace is still below limit, allocate more oldSpace"

    |limit free amount|

    limit := FreeSpaceGCLimit.
    limit notNil ifTrue:[
        "/ if reclaimed space is below limit, we have to allocate more
        "/ oldSpace, to avoid excessive gcSteps (due to freeSpaceLimit 
        "/ still not reached)
        "/
        free := self freeSpace + self freeListSpace.
        free < (limit * 3 // 2) ifTrue:[
            amount := FreeSpaceGCAmount.
            amount isNil ifTrue:[
                amount := limit * 3 // 2.
            ].
            'ObjectMemory [info]: moreOldSpace to satisfy free-limit' infoPrintCR.
            (self moreOldSpace:(amount - free + (64*1024))) ifFalse:[
                "/
                "/ could not increase oldspace; reset FreeSpaceGCLimit to avoid
                "/ useless retries
                'ObjectMemory [warning]: could not increase oldSpace - reset limit' errorPrintCR.
                FreeSpaceGCLimit := nil
            ]
        ].
    ].

    "Modified: 10.1.1997 / 17:59:48 / cg"
!

newSpaceSize:newSize
    "change the size of the newSpace. To do this, the current contents
     of the newSpace may have to be tenured (if size is smaller).
     Returns false, if it failed for any reason.
     Experimental: this interface may valish without notice.

     DANGER ALERT:
	be careful too big of a size may lead to longer scavenge pauses.
	Too small of a newSpace may lead to more CPU overhead, due to
	excessive scavenges. You have been warned."

%{  
    extern int __setNewSpaceSize();

    if (__isSmallInteger(newSize)) {
	RETURN (__setNewSpaceSize(__intVal(newSize)) ? true : false);
    }
%}.
    self primitiveFailed

    " less absolute CPU overhead (but longer pauses):

     ObjectMemory newSpaceSize:800*1024
    "

    " smaller pauses, but more overall CPU overhead:

     ObjectMemory newSpaceSize:200*1024
    "

    " the default:

     ObjectMemory newSpaceSize:400*1024
    "
!

oldSpaceCompressLimit
    "return the limit for oldSpace compression. If more memory than this
     limit is in use, the system will not perform compresses on the oldspace,
     but instead do a mark&sweep GC followed by an oldSpace increase if not enough
     could be reclaimed. The default is currently some 8Mb, which is ok for workstations
     with 16..32Mb of physical memory. If your system has much more physical RAM,
     you may want to increase this limit."

%{  /* NOCONTEXT */
    extern unsigned __compressingGCLimit();

    RETURN (__MKSMALLINT( __compressingGCLimit(-1) ));
%}
    "
     ObjectMemory oldSpaceCompressLimit
    "
!

oldSpaceCompressLimit:amount
    "set the limit for oldSpace compression. If more memory than this
     limit is in use, the system will not perform compresses on the oldspace,
     but instead do a mark&sweep GC followed by an oldSpace increase if not enough
     could be reclaimed. The default is currently some 8Mb, which is ok for workstations
     with 16..32Mb of physical memory. If your system has much more physical RAM,
     you may want to increase this limit.
     This method returns the previous increment value."

%{  /* NOCONTEXT */
    extern unsigned __compressingGCLimit();

    if (__isSmallInteger(amount)) {
	RETURN (__MKSMALLINT( __compressingGCLimit(__intVal(amount)) ));
    }
%}
    "to change the limit to 12Mb:"
    "
     ObjectMemory oldSpaceCompressLimit:12*1024*1024
    "
!

oldSpaceIncrement
    "return the oldSpaceIncrement value. Thats the amount by which
     more memory is allocated in case the oldSpace gets filled up.
     In normal situations, the default value used in the VM is fine
     and there is no need to change it."

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceIncrement();

    RETURN (__MKSMALLINT( __oldSpaceIncrement(-1) )); 
%}
    "
     ObjectMemory oldSpaceIncrement
    "
!

oldSpaceIncrement:amount
    "set the oldSpaceIncrement value. Thats the amount by which
     more memory is allocated in case the oldSpace gets filled up.
     In normal situations, the default value used in the VM is fine
     and there is no need to change it. This method returns the 
     previous increment value."

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceIncrement();

    if (__isSmallInteger(amount)) {
	RETURN (__MKSMALLINT( __oldSpaceIncrement(__intVal(amount)) )); 
    }
%}
    "to change increment to 1Meg:"
    "
     ObjectMemory oldSpaceIncrement:1024*1024
    "
!

tenureParameters:magic
    "this is pure magic and not for public eyes ...
     This method allows fine tuning the scavenger internals,
     in cooperation to some statistic & test programs.
     It is undocumented, secret and may vanish. 
     If you play around here, the system may behave very strange."

%{  /* NOCONTEXT */
    __tenureParams(magic);
%}.
!

turnGarbageCollectorOff
    "turn off the generational garbage collector by forcing new objects to be
     allocated directly in oldSpace (instead of newSpace)
     WARNING:
     This is somewhat dangerous: if collector is turned off,
     and too many objects are created, the system may run into trouble
     (i.e. oldSpace becomes full) and be forced to perform a full mark&sweep
     or even a compressing collect - making the overall realtime behavior worse.
     Use this only for special purposes or when realtime behavior
     is required for a limited time period.

     OBSOLETE: this is no longer supported 
	       - it may be a no-operation by the time you read this."

%{  /* NOCONTEXT */
    __allocForceSpace(OLDSPACE);
%}
!

turnGarbageCollectorOn
    "turn garbage collector on again (see ObjectMemory>>turnGarbageCollectorOff)"

%{  /* NOCONTEXT */
    __allocForceSpace(9999);
%}
!

watchTenure:flag
    "set/clear the tenureWatch. If set, an internalError exception will be raised,
     whenever objects are tenured from newSpace into oldSpace
     (except for an explicit tenure request). 
     This can be used to validate that no oldSpace objects are created
     (i.e. the system operates fully in newSpace). 
     Be careful, if the avoidTenure flag is not set, 
     there will almost always be a tenure sooner or later.

    EXPERIMENTAL - no warranty"

%{  /* NOCONTEXT */
    __watchTenure(flag == true ? 1 : 0);
%}
! !

!ObjectMemory class methodsFor:'interrupt handler access'!

childSignalInterruptHandler
    "return the handler for UNIX-death-of-a-childprocess-signal interrupts"

    ^ ChildSignalInterruptHandler
!

childSignalInterruptHandler:aHandler
    "set the handler for UNIX-death-of-a-childprocess-signal interrupts"

    ChildSignalInterruptHandler := aHandler

    "Created: 22.12.1995 / 14:14:52 / stefan"
    "Modified: 22.12.1995 / 14:15:16 / stefan"
!

customInterruptHandler
    "return the handler for custom interrupts"

    ^ CustomInterruptHandler
!

customInterruptHandler:aHandler
    "set the handler for custom interrupts"

    CustomInterruptHandler := aHandler
!

disposeInterruptHandler
    "return the handler for object disposal interrupts"

    ^ DisposeInterruptHandler
!

disposeInterruptHandler:aHandler
    "set the handler for object disposal interrupts"

    DisposeInterruptHandler := aHandler
!

errorInterruptHandler
    "return the handler for display error interrupts"

    ^ ErrorInterruptHandler
!

errorInterruptHandler:aHandler
    "set the handler for display error interrupts"

    ErrorInterruptHandler := aHandler
!

exceptionInterruptHandler
    "return the handler for floating point exception interrupts"

    ^ ExceptionInterruptHandler
!

internalErrorHandler
    "return the handler for ST/X internal errors.
     An internal error is reported for example when a methods
     bytecode is not a ByteArray, the selector table is not an Array
     etc.  
     Those should not occur in normal circumstances."

    ^ InternalErrorHandler
!

ioInterruptHandler
    "return the handler for I/O available signal interrupts (SIGIO/SIGPOLL)"

    ^ IOInterruptHandler
!

ioInterruptHandler:aHandler
    "set the handler for I/O available signal interrupts (SIGIO/SIGPOLL)"

    IOInterruptHandler := aHandler
!

recursionInterruptHandler
    "return the handler for recursion/stack overflow interrupts"

    ^ RecursionInterruptHandler
!

recursionInterruptHandler:aHandler
    "set the handler for recursion/stack overflow interrupts"

    RecursionInterruptHandler := aHandler
!

registerErrorInterruptHandler:aHandler forID:errorIDSymbol
    "register a handler"

    RegisteredErrorInterruptHandlers isNil ifTrue:[
	RegisteredErrorInterruptHandlers := IdentityDictionary new
    ].
    RegisteredErrorInterruptHandlers at:errorIDSymbol put:aHandler
!

registeredErrorInterruptHandlers
    "return registered handlers"

    ^ RegisteredErrorInterruptHandlers
!

signalInterruptHandler
    "return the handler for UNIX-signal interrupts"

    ^ SignalInterruptHandler
!

signalInterruptHandler:aHandler
    "set the handler for UNIX-signal interrupts"

    SignalInterruptHandler := aHandler
!

spyInterruptHandler
    "return the handler for spy-timer interrupts"

    ^ SpyInterruptHandler
!

spyInterruptHandler:aHandler
    "set the handler for spy-timer interrupts"

    SpyInterruptHandler := aHandler
!

stepInterruptHandler
    "return the handler for single step interrupts"

    ^ StepInterruptHandler
!

stepInterruptHandler:aHandler
    "set the handler for single step interrupts"

    StepInterruptHandler := aHandler
!

timerInterruptHandler
    "return the handler for timer interrupts"

    ^ TimerInterruptHandler
!

timerInterruptHandler:aHandler
    "set the handler for timer interrupts"

    TimerInterruptHandler := aHandler
!

userInterruptHandler
    "return the handler for CNTL-C interrupt handling"

    ^ UserInterruptHandler
!

userInterruptHandler:aHandler
    "set the handler for CNTL-C interrupt handling"

    UserInterruptHandler := aHandler
! !

!ObjectMemory class methodsFor:'interrupt statistics'!

interruptLatency:ms receiver:rec class:cls selector:sel vmActivity:vmActivity id:pid
    "example implementation of latencyTime monitoring:
     This method simply measures the max-latency time.
     You may want to use some other handler (see #interruptLatencyMonitor:)
     and extract more information (blocking context).
     DEMO Example."

    ms > MaxInterruptLatency ifTrue:[
        MaxInterruptLatency := ms.
        'IRQ-LATENCY: ' infoPrint. cls infoPrint. ' ' infoPrint. sel infoPrint. '(' infoPrint. vmActivity infoPrint . ') ---> ' infoPrint. ms infoPrintCR.
    ].
    (InterruptLatencyGoal notNil and:[ms > InterruptLatencyGoal]) ifTrue:[
        '*** IRQ REALTIME-DEADLINE MISSED: ' errorPrint. 
        cls errorPrint.
        ' ' errorPrint. sel errorPrint. '(' errorPrint. vmActivity errorPrint . ') ---> ' errorPrint. 
        ms errorPrintCR.
    ].

    "to enable the demo handler:

     ObjectMemory resetMaxInterruptLatency.
     ObjectMemory interruptLatencyMonitor:ObjectMemory.
    "
    "to disable timing statistics:

     ObjectMemory interruptLatencyMonitor:nil.
     ObjectMemory maxInterruptLatency printNL.
    "

    "Created: 7.11.1995 / 21:05:50 / cg"
    "Modified: 18.6.1996 / 14:15:52 / stefan"
    "Modified: 10.1.1997 / 19:09:53 / cg"
!

interruptLatencyGoal:millis
    "setup to report an error message, whenever a realtime goal could not be
     met due to blocked interrupts or long primitives or GC activity.
     An argument of nil clears the check.
     DEMO Example."

    InterruptLatencyGoal := millis.
    millis isNil ifTrue:[
	InterruptLatencyMonitor := nil.
    ] ifFalse:[
	MaxInterruptLatency := 0.
	InterruptLatencyMonitor := self.
    ]

    "
     ObjectMemory interruptLatencyGoal:50
    "
!

interruptLatencyMonitor
    "return the interrupt-latency-monitor if any. 
     See comment in #interruptLatencyMonitor:.
     This is a non-standard debugging/realtime instrumentation entry."

    ^ InterruptLatencyMonitor
!

interruptLatencyMonitor:aHandler
    "set the interrupt latency monitor. If non-nil, this one will be sent
     an interruptLatency: message with the millisecond delay between
     the interrupt and its handling.
     This is a non-standard debugging/realtime instrumentation entry."

    InterruptLatencyMonitor := aHandler
!

maxInterruptLatency
    "return the maximum accumulated interrupt latency in millis.
     DEMO Example."

    ^ MaxInterruptLatency
!

resetMaxInterruptLatency
    "reset the maximum accumulated interrupt latency probe time.
     DEMO Example."

    MaxInterruptLatency := 0
! !

!ObjectMemory class methodsFor:'just in time compilation'!

byteCodeSizeLimitForDynamicCompilation:aNumber
    "set a limit on a methods number of byteCodes.
     Compilation of a method into machine code is aborted,
     if its bytecode size is larger than the given number.
     This is only useful, if large methods have a smaller
     chance of being evaluated often (which may not be true).
     The predefined limit is some 4k (which seems to be ok)."

%{  /* NOCONTEXT */
    extern int __byteCodeSizeLimitForDynamicCompilation();
    int prev;

    prev = __byteCodeSizeLimitForDynamicCompilation(
                                    __isSmallInteger(aNumber)
                                    ? __intVal(aNumber)
                                    : -1);
    RETURN (__MKSMALLINT(prev));
%}.

    "
     ObjectMemory byteCodeSizeLimitForDynamicCompilation:nil
     ObjectMemory byteCodeSizeLimitForDynamicCompilation:8000
    "
!

codeForCPU:aCPUSymbol
    "instruct the just in time compiler that some specific CPU
     is present (forcing it to generate code for that).
     The default is set correct for the architecture, and the
     runTime system tries to figure out the cpu - but who knows ...
     The valid symbols depend on the architecture:
	i386:	#i486 / #i586		(affects code ordering)
	sparc:	#sparcV8 / #sparcV8	(affects instruction set)
	mips:	#rs2000 / #rs4000	(affects delay slot generation)

     This is a nonstandard entry and may be removed without notice - 
     not for the general user."

%{  /* NOCONTEXT */
    __codeForCPU(aCPUSymbol);
%}.

    "
     ObjectMemory codeForCPU:#i486
     ObjectMemory codeForCPU:#i586
    "
!

codeSizeLimitForDynamicCompilation:aNumber
    "set a limit on the resulting dynamic generates code
     size. Compilation of a method into machine code is aborted,
     if the resulting code is larger than the given number of
     bytes. This is only useful, if large methods have a smaller
     chance of being evaluated often (which may not be true).
     The predefined limit is some 4k (which seems to be ok)."

%{  /* NOCONTEXT */
    extern int __codeSizeLimitForDynamicCompilation();
    int prev;

    prev = __codeSizeLimitForDynamicCompilation( __isSmallInteger(aNumber)
				    ? __intVal(aNumber)
				    : -1);
    RETURN (__MKSMALLINT(prev));
%}.

    "
     ObjectMemory codeSizeLimitForDynamicCompilation:nil
     ObjectMemory codeSizeLimitForDynamicCompilation:8000
    "
!

compiledCodeCounter
    "return the number of additional code-bytes which
     were generated since the counter was last reset"

%{  /* NOCONTEXT */

    extern __compiledCodeCounter();
    int nBytes;

    nBytes = __compiledCodeCounter();
    RETURN (__MKSMALLINT(nBytes));
%}
    "
     ObjectMemory compiledCodeCounter
    "
!

compiledCodeSpaceUsed
    "return the actual number of bytes used for compiled code"

%{  /* NOCONTEXT */

    extern __compiledCodeSpaceUsed();
    int nBytes;

    nBytes = __compiledCodeSpaceUsed();
    RETURN (__MKSMALLINT(nBytes));
%}
    "
     ObjectMemory compiledCodeSpaceUsed
    "
!

fullSingleStepSupport
    "return the setting of the full single step support flag"

%{  /* NOCONTEXT */
    extern int __fullSingleStep();

    RETURN (__fullSingleStep(-1) ? true : false);
%}
    "
     ObjectMemory fullSingleStepSupport
    "

!

fullSingleStepSupport:aBoolean
    "enable/disable full single step support for the just-in-time-compiled code. 
     If off, things like simple increment/decrement, additions and variable-
     stores are not steppable, but treated like atomar invisible operations.
     If on, single step halts at those operations. 
     Execution is a bit slower if enabled."

%{  /* NOCONTEXT */
    extern int __fullSingleStep();
    int prev;

    prev = __fullSingleStep(aBoolean == true
                                   ? 1
                                   : (aBoolean == false)
                                        ? 0
                                        : -1);
    RETURN (prev ? true : false);
%}
    "
     ObjectMemory fullSingleStepSupport:true
     ObjectMemory fullSingleStepSupport:false
    "

!

getCompiledCodeLimit
    "get the codeLimit from the VM"

%{  /* NOCONTEXT */
    extern int __dynamicCodeLimit();
    int limit;

    limit = __dynamicCodeLimit();
    if (limit) {
        RETURN (__MKSMALLINT(limit));
    }
    RETURN (nil);
%}
!

insnSizeLimitForDynamicCompilation:aNumber
    "set a limit on a methods number of internal insns.
     Compilation of a method into machine code is aborted,
     if during compilation, more than the given number of
     internal insns are generated.
     The limit controls the amount of dynamic memory allocated
     during compilation and may be changed for small-memory
     systems.
     The predefined limit is some 4k (which seems to be ok)."

%{  /* NOCONTEXT */
    extern int __insnSizeLimitForDynamicCompilation();
    int prev;

    prev = __insnSizeLimitForDynamicCompilation( __isSmallInteger(aNumber)
                                    ? __intVal(aNumber)
                                    : -1);
    RETURN (__MKSMALLINT(prev));
%}.

    "
     ObjectMemory insnSizeLimitForDynamicCompilation:nil
     ObjectMemory insnSizeLimitForDynamicCompilation:8000
    "
!

justInTimeCompilation
    "return the value of the just-in-time-compilation flag"

%{  /* NOCONTEXT */
    extern int __justInTimeCompilation();

    RETURN (__justInTimeCompilation(-1) ? true : false);
%}
    "
     ObjectMemory justInTimeCompilation
    "

!

justInTimeCompilation:aBoolean
    "enable/disable just-in-time-compilation."

    aBoolean notNil ifTrue:[
        JustInTimeCompilationEnabled := aBoolean.
    ].

%{  /* NOCONTEXT */
    extern int __justInTimeCompilation();
    int prev;

    prev = __justInTimeCompilation(aBoolean == true
                                   ? 1
                                   : (aBoolean == false)
                                        ? 0
                                        : -1);
    RETURN (prev ? true : false);
%}.
    "
     ObjectMemory justInTimeCompilation:true
     ObjectMemory justInTimeCompilation:false
    "
!

reEnableJustInTimeCompilation
    "to be called after a snapshot restart; if justInTimeCompiler
     was enabled before, do it again. 
     For now, this is not done automatically, to allow restarting
     a system with the dynamic compiler turned off (its still experimental).
     Therefore, this reenabling is done in the smalltalk_r.rc file."

    JustInTimeCompilationEnabled == true ifTrue:[
	self justInTimeCompilation:true
    ]
!

resetCompiledCodeCounter
    "reset the counter of additional code-bytes"

%{  /* NOCONTEXT */

    extern void __resetDynamicCodeGeneratedCounter();

    __resetDynamicCodeGeneratedCounter();
%}
    "
     ObjectMemory resetCompiledCodeCounter
    "

!

setCompiledCodeLimit:newLimit
    "set the VM's limit"

%{  /* NOCONTEXT */
    extern __setDynamicCodeLimit();

    if (__isSmallInteger(newLimit)) {
	__setDynamicCodeLimit(__intVal(newLimit));
    } else if (newLimit == nil) {
	__setDynamicCodeLimit(0);
    }
%}
!

supportsJustInTimeCompilation
    "return true, if this system supports just-in-time-compilation of
     bytecode to machine code. Dont confuse this with stc-compilation."

%{  /* NOCONTEXT */
    extern int __canDoJustInTimeCompilation();

    RETURN (__canDoJustInTimeCompilation() ? true : false);
%}
    "
     ObjectMemory supportsJustInTimeCompilation
    "

! !

!ObjectMemory class methodsFor:'low memory handling'!

memoryInterrupt
    "when a low-memory condition arises, ask all classes to
     remove possibly cached data. You may help the system a bit,
     in providing a lowSpaceCleanup method in your classes which have
     lots of data kept somewhere (usually, cached data).
     - this may or may not help."

    self performLowSpaceCleanup.
"/    self error:'almost out of memory'
    'ObjectMemory [warning]: almost out of memory' errorPrintCR.

    LowSpaceSemaphore signalIf.

    "Modified: 10.1.1997 / 17:59:31 / cg"
!

performLowSpaceCleanup
    "ask all classes to remove possibly cached data. 
     You may help the system a bit, in providing a lowSpaceCleanup method 
     in your classes which have lots of data kept somewhere (usually, cached data)."

    Smalltalk allBehaviorsDo:[:aClass |
        aClass lowSpaceCleanup
    ].

    "Created: 12.4.1996 / 14:57:28 / cg"
! !

!ObjectMemory class methodsFor:'object finalization'!

allChangedShadowObjectsDo:aBlock
    "evaluate the argument, aBlock for all known shadow objects which have
     lost a pointer recently."
%{
    __allChangedShadowObjectsDo(&aBlock);
%}
!

disposeInterrupt
    "this is triggered by the garbage collector,
     whenever any shadowArray looses a pointer."

    FinalizationSemaphore notNil ifTrue:[
        "/
        "/ background finalizer is waiting ...
        "/
        FinalizationSemaphore signalOnce
    ] ifFalse:[
        "/
        "/ do it right here
        "/
        self finalize
    ]

    "Modified: 13.1.1997 / 17:26:16 / cg"
!

finalize
    "tell all weak objects that something happened."

    self allChangedShadowObjectsDo:[:aShadowArray | 
        aShadowArray lostPointer.
    ].

    "/ we should change the setup,
    "/ to make the Dependencies collection a dependent
    "/ of all the WeakArrays and WeakIDSets there,
    "/ and send a changed message when any of them looses
    "/ a pointer. 
    "/ This would automize the send below.
    "/
    "/ WARNING:
    "/   this can only be done, if the WeakIDSet holds its
    "/   dependends itself (like a WeakArray), 
    "/   otherwise we might get trouble.
    "/   Therefore, things are as they are now ;-)

    Dependencies removeEmptyDependencyValues

    "Modified: 9.1.1997 / 01:00:44 / cg"
!

startBackgroundFinalizationAt:aPriority
    "start a process doing finalization work in the background.
     Can be used to reduce the pauses created by finalization.
     Normally, these pauses are not noticed; however if you have (say)
     ten thousands of weak objects, these could become long enough to
     make background finalization usefull.
     WARNING: background finalization may lead to much delayed freeing of
     system resources. Especially, you may temporarily run out of free
     color table entries or fileDescriptors etc. Use at your own risk (if at all)"

    |p|

    "/
    "/ its not useful, to run it more than once
    "/
    BackgroundFinalizationProcess notNil ifTrue:[
	BackgroundFinalizationProcess priority:aPriority.
	^ self
    ].

    FinalizationSemaphore := Semaphore new name:'FinalizationSemaphore'.

    p :=
	[
	    [
		[true] whileTrue:[
		    "
		     wait till something to do ...
		    "
		    FinalizationSemaphore wait.
		    "
		     ... and do it
		    "
		    self finalize
		]
	    ] valueOnUnwindDo:[
		BackgroundFinalizationProcess := nil.
		FinalizationSemaphore := nil
	    ]
	] newProcess.
    p name:'background finalizer'.
    p priority:aPriority.
    p restartable:true.
    p resume.
    BackgroundFinalizationProcess := p

    "
     ObjectMemory startBackgroundFinalizationAt:5
    "
!

stopBackgroundFinalization
    "stop the background finalizer"

    BackgroundFinalizationProcess notNil ifTrue:[
	BackgroundFinalizationProcess terminate.
	BackgroundFinalizationProcess := nil
    ].

    "
     ObjectMemory stopBackgroundFinalization
    "
! !

!ObjectMemory class methodsFor:'physical memory access'!

collectedOldSpacePagesDo:aBlock
    "evaluates aBlock for all pages in the prev. oldSpace, passing
     the pages address as argument. 
     For internal & debugging use only."
%{
    if (__collectedOldSpacePagesDo(&aBlock) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

newSpacePagesDo:aBlock
    "evaluates aBlock for all pages in the newSpace, passing
     the pages address as argument.
     For internal & debugging use only."
%{
    if (__newSpacePagesDo(&aBlock) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

oldSpacePagesDo:aBlock
    "evaluates aBlock for all pages in the oldSpace, passing
     the pages address as argument. 
     For internal & debugging use only."
%{
    if (__oldSpacePagesDo(&aBlock) < 0) {
	RETURN (false);
    }
%}.
    ^ true
!

pageIsInCore:aPageNumber
    "return true, if the page (as enumerated via oldSpacePagesDo:)
     is in memory; false, if currently paged out. For internal
     use / monitors only; may vanish.
     NOTICE: not all systems provide this information; on those that
     do not, true is returned for all pages."
%{
#ifdef HAS_MINCORE
    int pageSize = getpagesize();
    char result[10];
    INT addr;

    if (__isSmallInteger(aPageNumber)) {
	addr = __intVal(aPageNumber) & ~(pageSize - 1);
    } else {
	addr = ((INT)aPageNumber) & ~(pageSize - 1);
    }
    if (mincore(addr, pageSize-1, result) < 0) {
	RETURN (true);
    }
    RETURN ((result[0] & 1) ? true : false);
#endif
%}.
    "OS does not supply this info - assume yes"
    ^ true
! !

!ObjectMemory class methodsFor:'queries'!

bytesUsed
    "return the number of bytes allocated for objects -
     this number is not exact, since some objects may already be dead
     (i.e. not yet reclaimed by the garbage collector).
     If you need the exact number, you have to loop over all
     objects and ask for the bytesize using ObjectMemory>>sizeOf:."

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceUsed(), __newSpaceUsed(), __freeListSpace();

    RETURN ( __MKSMALLINT(__oldSpaceUsed() + __newSpaceUsed() - __freeListSpace()) );
%}
    "
     ObjectMemory bytesUsed  
    "
!

collectObjectsWhich:aBlock
    "helper for the whoReferences queries. Returns a collection
     of objects for which aBlock returns true."

    |aCollection|

    aCollection := IdentitySet new.
    self allObjectsDo:[:o |
	(aBlock value:o) ifTrue:[
	    aCollection add:o
	]
    ].
    (aCollection size == 0) ifTrue:[
	"actually this cannot happen - there is always one"
	^ nil
    ].
    ^ aCollection
!

fixSpaceSize
    "return the total size of the fix space."

%{  /* NOCONTEXT */
    extern unsigned __fixSpaceSize();

    RETURN ( __MKSMALLINT(__fixSpaceSize()) );
%}
    "
     ObjectMemory fixSpaceSize
    "
!

fixSpaceUsed
    "return the number of bytes allocated for old objects in fix space."

%{  /* NOCONTEXT */
    extern unsigned __fixSpaceUsed();

    RETURN ( __MKSMALLINT(__fixSpaceUsed()) );
%}
    "
     ObjectMemory fixSpaceUsed
    "
!

freeListSpace
    "return the number of bytes in the free lists.
     (which is included in oldSpaceUsed)"

%{  /* NOCONTEXT */
    extern unsigned __freeListSpace();

    RETURN ( __MKSMALLINT(__freeListSpace()) );
%}
    "
     ObjectMemory freeListSpace
    "
!

freeSpace
    "return the number of bytes in the compact free area.
     (oldSpaceUsed + freeSpaceSize = oldSpaceSize)"

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceSize(), __oldSpaceUsed();

    RETURN ( __MKSMALLINT(__oldSpaceSize() - __oldSpaceUsed()) );
%}
    "
     ObjectMemory freeSpace
    "
!

garbageCollectCount
    "return the number of compressing collects that occurred since startup"

%{  /* NOCONTEXT */
    extern int __garbageCollectCount();

    RETURN (__MKSMALLINT(__garbageCollectCount()));
%}
    "
     ObjectMemory garbageCollectCount 
    "
!

incrementalGCCount
    "return the number of incremental collects that occurred since startup"

%{  /* NOCONTEXT */
    extern int __incrementalGCCount();

    RETURN (__MKSMALLINT(__incrementalGCCount()));
%}
    "
     ObjectMemory incrementalGCCount
    "
!

incrementalGCPhase
    "returns the internal state of the incremental GC.
     The meaning of those numbers is a secret :-).
     (for the curious: (currently)
      2 is idle, 3..11 are various mark phases,
      12 is the sweep phase. 0 and 1 are cleanup phases when the
      incr. GC gets interrupted by a full GC).
     Do not depend on the values - there may be additional phases in
     future versions (incremental compact ;-).
     This is for debugging and monitoring only - and may change or vanish"

%{  /* NOCONTEXT */
    extern int __incrGCphase();

    RETURN (__MKSMALLINT(__incrGCphase()));
%}
!

lastScavengeReclamation
    "returns the number of bytes replacimed by the last scavenge.
     For statistic only - this may vanish."

%{  /* NOCONTEXT */
    extern int __newSpaceReclaimed();

    RETURN ( __MKSMALLINT(__newSpaceReclaimed()) );
%}
    "percentage of reclaimed objects is returned by:

     ((ObjectMemory lastScavengeReclamation)
      / (ObjectMemory newSpaceSize)) * 100.0  
    "
!

lifoRememberedSet
    "return the lifoRemSet.
     This is pure VM debugging and will vanish without notice."

%{  /* NOCONTEXT */
    extern OBJ __lifoRememberedSet();

    RETURN ( __lifoRememberedSet() );
%}
    "
     ObjectMemory lifoRememberedSet
    "
!

lifoRememberedSetSize
    "return the size of the lifoRemSet.
     This is a VM debugging interface and may vanish without notice."

%{  /* NOCONTEXT */
    extern int __lifoRememberedSetSize();

    RETURN (__MKSMALLINT(__lifoRememberedSetSize()));
%}
    "
     ObjectMemory lifoRememberedSetSize
    "
!

markAndSweepCount
    "return the number of mark&sweep collects that occurred since startup"

%{  /* NOCONTEXT */
    extern int __markAndSweepCount();

    RETURN (__MKSMALLINT(__markAndSweepCount()));
%}
    "
     ObjectMemory markAndSweepCount 
    "
!

maximumIdentityHashValue
    "for ST-80 compatibility: return the maximum value
     a hashKey as returned by identityHash can get.
     Since ST/X uses direct pointers, a field in the objectHeader
     is used, which is currently 11 bits in size."

%{  /* NOCONTEXT */
    RETURN ( __MKSMALLINT( __MAX_HASH__ << __HASH_SHIFT__) );
%}
    "
     ObjectMemory maximumIdentityHashValue
    "
!

minScavengeReclamation
    "returns the number of bytes replacimed by the least effective scavenge.
     For statistic only - this may vanish."

%{  /* NOCONTEXT */
    extern int __newSpaceReclaimedMin();

    RETURN ( __MKSMALLINT(__newSpaceReclaimedMin()) );
%}
    "
     ObjectMemory minScavengeReclamation
    "
!

newSpaceSize
    "return the total size of the new space - this is usually fix"

%{  /* NOCONTEXT */
    extern unsigned __newSpaceSize();

    RETURN ( __MKSMALLINT(__newSpaceSize()) );
%}
    "
     ObjectMemory newSpaceSize
    "
!

newSpaceUsed
    "return the number of bytes allocated for new objects.
     The returned value is usually obsolete as soon as you do
     something with it ..."

%{  /* NOCONTEXT */
    extern unsigned __newSpaceUsed();

    RETURN ( __MKSMALLINT(__newSpaceUsed()) );
%}
    "
     ObjectMemory newSpaceUsed   
    "
!

numberOfObjects
    "return the number of objects in the system."

    |tally "{ Class: SmallInteger }"|

    tally := 0.
    self allObjectsDo:[:obj | tally := tally + 1].
    ^ tally

    "
     ObjectMemory numberOfObjects  
    "
!

numberOfWeakObjects
    "return the number of weak objects in the system"

%{  /* NOCONTEXT */
    extern int __weakListSize();

    RETURN ( __MKSMALLINT(__weakListSize()) );
%}
    "
     ObjectMemory numberOfWeakObjects
    "
!

oldSpaceAllocatedSinceLastGC
    "return the number of bytes allocated for old objects since the
     last oldspace garbage collect occured. This information is used
     by ProcessorScheduler to decide when to start the incremental
     background GC."

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceAllocatedSinceLastGC();

    RETURN ( __MKSMALLINT(__oldSpaceAllocatedSinceLastGC()) );
%}
    "
     ObjectMemory oldSpaceAllocatedSinceLastGC   
    "
!

oldSpaceSize
    "return the total size of the old space. - may grow slowly"

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceSize();

    RETURN ( __MKSMALLINT(__oldSpaceSize()) );
%}
    "
     ObjectMemory oldSpaceSize
    "
!

oldSpaceUsed
    "return the number of bytes allocated for old objects.
     (This includes the free lists)"

%{  /* NOCONTEXT */
    extern unsigned __oldSpaceUsed();

    RETURN ( __MKSMALLINT(__oldSpaceUsed()) );
%}
    "
     ObjectMemory oldSpaceUsed  
    "
!

rememberedSetSize
    "return the number of old objects referencing new ones.
     This is a VM debugging interface and may vanish without notice."

%{  /* NOCONTEXT */
    extern int __rememberedSetSize();

    RETURN (__MKSMALLINT(__rememberedSetSize()));
%}
    "
     ObjectMemory rememberedSetSize
    "
!

resetMinScavengeReclamation
    "resets the number of bytes replacimed by the least effective scavenge.
     For statistic only - this may vanish."

%{  /* NOCONTEXT */
    extern int __resetNewSpaceReclaimedMin();

    __resetNewSpaceReclaimedMin();
%}.
    ^ self
    "
     ObjectMemory resetMinScavengeReclamation.
     ObjectMemory minScavengeReclamation
    "
!

runsSingleOldSpace
    "return true, if the system runs in a single oldSpace or
     false if not.
     The memory system will always drop the second semispace when 
     running out of virtual memory, or the baker-limit is reached.
     OBSOLETE: 
	 the system may now decide at any time to switch between
	 single and double-space algorithms, depending on the overall memory
	 size. You will now almost always get false as result, since the
	 second semispace is only allocated when needed, and released
	 immediately afterwards.
    "

%{  /* NOCONTEXT */
    extern int __runsSingleOldSpace();

    RETURN ( (__runsSingleOldSpace() ? true : false) );
%}
    "
     ObjectMemory runsSingleOldSpace 
    "
!

scavengeCount
    "return the number of scavenges that occurred since startup"

%{  /* NOCONTEXT */
    extern int __scavengeCount();

    RETURN (__MKSMALLINT(__scavengeCount()));
%}
    "
     ObjectMemory scavengeCount 
    "
!

symSpaceSize
    "return the total size of the sym space."

%{  /* NOCONTEXT */
    extern unsigned __symSpaceSize();

    RETURN ( __MKSMALLINT(__symSpaceSize()) );
%}
    "
     ObjectMemory symSpaceSize
    "
!

symSpaceUsed
    "return the number of bytes allocated for old objects in sym space."

%{  /* NOCONTEXT */
    extern unsigned __symSpaceUsed();

    RETURN ( __MKSMALLINT(__symSpaceUsed()) );
%}
    "
     ObjectMemory symSpaceUsed
    "
!

tenureAge
    "return the current tenure age - thats the number of times
     an object has to survive scavenges to be moved into oldSpace.
     For statistic/debugging only - this method may vanish"

%{  /* NOCONTEXT */
    extern unsigned __tenureAge();

    RETURN ( __MKSMALLINT(__tenureAge()) );
%}
!

vmSymbols
    "return a collection of symbols used by the VM"

    "/ DO NOT REMOVE THIS METHOD.
    "/ It is required to force the compiler to create those symbols
    "/ in fixed (non-moving) memory.

    ^ #(
      "/ interpreter stuff

      "/ JavaVM stuff

      nativeMethodInvokation
      newCleared
      monitorEnter:
      monitorExit:
      classInit
      javaClass
      resolve
      resolveStatic
    )
!

whoReferences:anObject
    "return a collection of objects referencing the argument, anObject"

    ^ self collectObjectsWhich:[:o | o references:anObject]

    "
     (ObjectMemory whoReferences:Transcript) printNL
    "
!

whoReferencesDerivedInstancesOf:aClass
    "return a collection of objects refering to instances
     of the argument, aClass or a subclass of it."

    ^ self collectObjectsWhich:[:o | o referencesDerivedInstanceOf:aClass]

    "
     (ObjectMemory whoReferencesDerivedInstancesOf:View) printNL
    "
!

whoReferencesInstancesOf:aClass
    "return a collection of objects refering to instances
     of the argument, aClass"

    ^ self collectObjectsWhich:[:o | o referencesInstanceOf:aClass]

    "
     (ObjectMemory whoReferencesInstancesOf:SystemBrowser) printNL
    "
! !

!ObjectMemory class methodsFor:'semaphore access'!

lowSpaceSemaphore
    "return the semaphore that is signalled when the system detects a
     low space condition. Usually, some time after this, an allocationFailure
     will happen. You can have a cleanup process sitting in that semaphore and
     start to release object."

    ^ LowSpaceSemaphore
! !

!ObjectMemory class methodsFor:'statistics'!

ageStatistic
    "for ST/X developers only:
     dump contents of newSpace with objects age information.
     This method may be removed without notice"

%{   /* NOCONTEXT */

    __ageStatistics();
%}
    "
     ObjectMemory ageStatistic
    "
!

codeCacheInfo
    "for ST/X developers only:
     dump contents of dynamic code cache LRU lists.
     This method may be removed without notice"

%{   /* NOCONTEXT */
    OBJ __codeCacheInfo();

    RETURN (__codeCacheInfo());
%}
    "
     ObjectMemory codeCacheInfo do:[:item |
        |n nMethods nBytes|

        n := item at:1.
        nMethods := item at:2.
        nBytes := item at:3.

        n isNil ifTrue:[
            '>>' print
        ] ifFalse:[
            (n printStringLeftPaddedTo:2) print.
        ].
        ' ' print.
        (nMethods printStringLeftPaddedTo:4) print.
        ' ' print.
        (nBytes printStringLeftPaddedTo:6) printCR.
     ]
    "

! !

!ObjectMemory class methodsFor:'system management'!

allBinaryModulesDo:aBlock
    "internal private method - walk over all known binary
     modules and evaluate aBlock for each entry.
     Do not depend on the information returned for each - this may
     change without notice."

%{
    __REGISTRATION_DO_BLOCK(&aBlock COMMA_SND);
%}
!

binaryModuleInfo
    "return a collection of moduleInfo entries.
     This returns a dictionary (keys are internal moduleIDs)
     with one entry for each binary package (module)."

    |modules|

    modules := IdentityDictionary new.
    self allBinaryModulesDo:[:entry | 
        |id name type libName subModuleName module dynamic infoRec pathName
         typeName nameString|

        id := entry at:1.
        subModuleName := (entry at:2) asSymbol.
        libName := (entry at:4).

        id > 0 ifTrue:[
            pathName := ObjectFileLoader pathNameFromID:id.
            dynamic := true.
            typeName := 'dynamic '.
            pathName isNil ifTrue:[
                name := '?'
            ] ifFalse:[
                name := pathName asFilename baseName
            ]
        ] ifFalse:[
            dynamic := false.
            typeName := 'builtIn '.
            pathName := nil.
            libName isNil ifTrue:[
                name := subModuleName
            ] ifFalse:[
                name := libName
            ].
        ].
        nameString := typeName.
        libName isNil ifTrue:[
            nameString := nameString, 'module '
        ] ifFalse:[
            nameString := nameString, 'classLib '
        ].
        nameString := nameString , name.

        libName isNil ifTrue:[
            type := #classObject
        ] ifFalse:[
            type := #classLibrary
        ].

        infoRec := modules at:id ifAbsent:nil.
        infoRec notNil ifTrue:[
            infoRec classNames add:subModuleName.
        ] ifFalse:[
            infoRec := BinaryModuleDescriptor 
                            name:nameString 
                            type:type 
                            id:id 
                            dynamic:dynamic 
                            classNames:(Set with:subModuleName) 
                            pathName:pathName 
                            libraryName:libName 
                            timeStamp:nil.

            modules at:id put:infoRec.
        ].
    ].
    ^ modules

    "
     ObjectMemory binaryModuleInfo
    "

    "Modified: 17.9.1995 / 16:33:02 / claus"
    "Modified: 22.4.1997 / 23:42:59 / cg"
!

fullBinaryModuleInfo
    "return a full collection of moduleInfo entries.
     This returns a dictionary (keys are component names)
     with one entry for each sub-component in all binary packages."

    |modules|

    modules := IdentityDictionary new.
    self allBinaryModulesDo:[:entry | 
        |infoEntry
         id name type libName subModuleName module dynamic pathName t|

        id := entry at:1.
        subModuleName := (entry at:2) asSymbol.
        libName := (entry at:4).
        t := AbsoluteTime fromOSTime:((entry at:5) * 1000).

        id > 0 ifTrue:[
            dynamic := true.
        ] ifFalse:[
            dynamic := false.
        ].
        libName isNil ifTrue:[
            type := #classObject
        ] ifFalse:[
            type := #classLibrary
        ].

        "/ ... new:
        infoEntry := BinaryModuleDescriptor 
                        name:nil 
                        type:type 
                        id:id 
                        dynamic:dynamic 
                        classNames:(Array with:subModuleName) 
                        pathName:nil 
                        libraryName:libName 
                        timeStamp:t.

        modules at:subModuleName put:infoEntry.
    ].
    ^ modules

    "
     ObjectMemory fullBinaryModuleInfo
    "

    "Modified: 30.8.1995 / 17:29:30 / claus"
    "Modified: 31.10.1996 / 13:58:44 / cg"
!

imageBaseName
    "return a reasonable filename to use as baseName (i.e. without extension).
     This is the filename of the current image (without '.img') or,
     if not running from an image, the default name 'st'"

    |nm|

    nm := ImageName.
    (nm isNil or:[nm isBlank]) ifTrue:[
	^ 'st'
    ].
    (nm endsWith:'.sav') ifTrue:[
	nm := nm copyWithoutLast:4
    ].
    (nm endsWith:'.img') ifTrue:[
	^ nm copyWithoutLast:4
    ].
    ^ nm

    "
     ObjectMemory imageBaseName    
    "
!

imageName
    "return the filename of the current image, or nil
     if not running from an image."

    ^ ImageName

    "
     ObjectMemory imageName 
    "
!

imageSaveTime
    "return a timestamp for when the running image was saved.
     Return nil if not running from an image."

    ^ ImageSaveTime
!

loadClassBinary:aClassName
    "find the object file for aClassName and -if found - load it;
     this one loads precompiled object files"

    |fName newClass|

    fName := self fileNameForClass:aClassName.
    fName notNil ifTrue:[
	Class withoutUpdatingChangesDo:
	[
	    self loadBinary:(fName , '.o')
	].
	newClass := self at:(aClassName asSymbol).
	(newClass notNil and:[newClass implements:#initialize]) ifTrue:[
	    newClass initialize
	]
    ]
!

nameForChanges
    "return a reasonable filename to store the changes into.
     Currently, this is defined in a classVariable and defaults to 'changes'.
     In future versions, this will be the basename of the current image with '.img' replaced
     by '.chg', or, if not running from an image, the default name 'st.chg'."

    ChangeFileName notNil ifTrue:[^ ChangeFileName].
    ^ 'changes'.

"/ future versions will have:
"/ (requires some additionas at other places)
"/
"/    ^ self imageBaseName , '.chg'

    "
     ObjectMemory nameForChanges    
    "
!

nameForChanges:aFilename
    "set the name of the file where changes are stored into."

    ChangeFileName := aFilename

    "
     ObjectMemory nameForChanges:'myChanges'    
    "
!

nameForSnapshot
    "return a reasonable filename to store the snapshot image into.
     This is the filename of the current image or,
     if not running from an image, the default name 'st.img'"

    ^ self imageBaseName , '.img'

    "
     ObjectMemory nameForSnapshot    
    "
!

nameForSources
    "return a reasonable filename to store the sources into.
     This is the basename of the current image with '.img' replaced
     by '.src', or, if not running from an image, the default name 'st.src'"

    ^ self imageBaseName , '.src'

    "
     ObjectMemory nameForSources    
    "
!

primSnapShotOn:aFileName
    "create a snapshot in the given file.
     Low level entry. Does not notify classes or write an entry to
     the changes file. Also, no image backup is created. 
     Returns true if the snapshot worked, false if it failed for some reason.
     This method should not be used in normal cases."

    |ok oldImageName oldImageTime|

    "
     save the name and time with it ...
    "
    oldImageName := ImageName.
    oldImageTime := ImageSaveTime.

    ImageName := aFileName.
    ImageSaveTime := AbsoluteTime now.

%{  /* CALLSSTACK:32000 */

    OBJ __snapShotOn();
    OBJ funny = @symbol(funnySnapshotSymbol);

    if (__isString(aFileName)) {
	__BLOCKINTERRUPTS();
	ok = __snapShotOn(__context, _stringVal(aFileName), funny);
	__UNBLOCKINTERRUPTS();
    }
%}.

    ImageName := oldImageName.
    ImageSaveTime := oldImageTime.

    ^ ok
!

snapShot
    "create a snapshot file containing all of the current state."

    self snapShotOn:(self nameForSnapshot)

    "
     ObjectMemory snapShot
    "
!

snapShotOn:aFileName
    "create a snapshot in the given file. 
     If the file exists, save it for backup.
     Return true if the snapshot worked, false if it failed for some reason.
     Notify dependents before and after the snapshot operation."

    |ok fn|

    "
     keep a save version - just in case something
     bad happens while writing the image.
     (could be st/x internal error or file-system errors etc)
    "
    fn := aFileName asFilename.
    fn exists ifTrue:[
	fn renameTo:(fn withSuffix:'sav')
    ].

    "
     give others a chance to fix things
    "
    self changed:#save.             "/ will vanish ...
    self changed:#aboutToSnapshot.  "/ ... for ST-80 compatibility

    "
     ST-80 compatibility; send #preSnapshot to all classes
    "
    Smalltalk allBehaviorsDo:[:aClass |
	aClass preSnapshot
    ].

    ok := self primSnapShotOn:aFileName.

    ok ifTrue:[
	Class addChangeRecordForSnapshot:aFileName.
    ].


    "
     ST-80 compatibility; send #postSnapshot to all classes
    "
    Smalltalk allBehaviorsDo:[:aClass |
	aClass postSnapshot
    ].
    self changed:#finishedSnapshot.  "/ ST-80 compatibility
    ^ ok

    "
     ObjectMemory snapShotOn:'myimage.img'
    "
! !

!ObjectMemory::BinaryModuleDescriptor class methodsFor:'instance creation'!

name:n type:t id:i dynamic:d classNames:c pathName:p libraryName:l timeStamp:ts
    ^ (self basicNew) name:n type:t id:i dynamic:d classNames:c pathName:p libraryName:l timeStamp:ts


! !

!ObjectMemory::BinaryModuleDescriptor methodsFor:'accessing'!

classNames
    "return the names of the classes contained in that module"

    ^ classNames
!

dynamic
    "return true, if this module was loaded dynamically 
     (as opposed to a module which was linked into the system right from
      the start. 
      Do not cofuse this with shared libraries - these may be dynamic or not)."

    ^ dynamic
!

id
    "return the modules internal id; this is (currently) a small number,
     with positive values for dynamically added modules.
     Modules which were present at startup (both statically linked or shared libraries)
     have a negative id (Do not depend on this - use #dynamic instead)."

    ^ id
!

libraryName
    "return the name of this library. Typically, this is the libraries
     filename without a suffix (such as 'libbasic')"

    ^ libraryName
!

name
    "return the modules name - usually, some descriptive text plus the modules
     libraryName."

    ^ name
!

pathName
    "return the modules pathName"

    ^ pathName
!

timeStamp
    "return the modules timeStamp"

    ^ timeStamp
!

type
    "return the modules type. This is a symbol, such as #classLibrary."

    ^ type
! !

!ObjectMemory::BinaryModuleDescriptor methodsFor:'backward compatibility'!

at:key
    "backward compatibility access: in previous releases, IdentityDictionaries
     were used to hold my information. Allow access via key messages.
     This method will vanish - use the proper access protocol."

    ^ self perform:key
! !

!ObjectMemory::BinaryModuleDescriptor methodsFor:'private access'!

name:n type:t id:i dynamic:d classNames:c pathName:p libraryName:l timeStamp:ts
    name := n.
    type := t.
    id := i.
    dynamic := d.
    classNames := c.
    libraryName := l.
    timeStamp := ts.
    pathName := p. 
! !

!ObjectMemory class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Attic/ObjMem.st,v 1.151 1998-01-26 11:59:17 cg Exp $'
! !
ObjectMemory initialize!
